
/* =========================================================================== *
|  
|  Copyright (c) by the UA computer vision group (representing affiliated
|  authors and institutions).  
|  
|  Authors of this code include:
|      Prasad Gabbur
|      Kobus Barnard
| 
|  This code is under development and is currently restricted to use within the
|  UA computer vision group except when granted on a case by case basis.
|  (Granting of permission is implicit if a vision group member explicitly
|  provides access to the code, perhaps by E-mailing a tar file). Permission for
|  further distribution of the code is not granted.  Permission for use is
|  further subject to the following general conditions for distribution of UA
|  computer vision code.
| 
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowledged in publications, and relevant papers are cited.
| 
|  For other use contact kobus AT cs DOT arizona DOT edu.
| 
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarantee of performance, support, or
|  fitness for any particular task. Nonetheless, we are interested in hearing
|  about problems that you encounter. 
|
*  ========================================================================== */

#include "m/m_incl.h"      /*  Only safe if first #include in a ".c" file  */

#include "n/n_invert.h"  
#include "n/n_svd.h"  

#include "r/r_cluster.h"  
#include "r/r_cluster_lib.h"  

/* Some C++ compilers need this. (not g++). */
#ifdef __cplusplus
extern "C" {
#endif

/* -------------------------------------------------------------------------- */

/*
 * Gauss routine slows things down a bit--used for debugging.
 * 
*/
#define DONT_USE_GAUSS_ROUTINE

#define EM_MIN_CLUSTER_SIZE  (1e-10)
#define REQUIRED_SUBSET_FIT  (1e-20)

#define USE_LOGS
/*
#define SHORT_CUT_POINTERS
*/

/* Question--does this affect the scaling of output parameters? If so, it could
// screw up searches! */
/* #define RESCALE  */

#define DONT_PLOT_GRAD_BUT_DO_GRAD_ASCENT  
    
/* -------------------------------------------------------------------------- */

static double fs_data_perturbation                           = DBL_NOT_SET;
static int    fs_plot_log_likelihood_vs_num_clusters         = FALSE;
static int    fs_num_tries_per_cluster_count                 = 1;
static int    fs_num_cluster_count_samples                   = 30;
static int    fs_max_num_iterations                          = 20;
static double fs_iteration_tolerance                         = 1.0e-6;
static double fs_dis_item_prob_threshold                     = DBL_NOT_SET;
static int    fs_normalize_data                              = FALSE;    /* Switched Dec 21, 2004. */
static double fs_var_offset                                  = 0.0001;
static int    fs_use_unbiased_var_estimate_in_M_step         = FALSE;
static double fs_held_out_data_fraction                      = 0.1;
static int    fs_tie_var                                     = FALSE;
static int    fs_tie_feature_var                             = FALSE;
static int    fs_tie_cluster_var                             = FALSE;
static int    fs_model_selection_training_MDL                = FALSE;
static int    fs_model_selection_held_out_LL                 = TRUE;
static int    fs_model_selection_held_out_MDL                = FALSE;
static int    fs_model_selection_held_out_corr_diff          = TRUE;
static int    fs_model_selection_held_out_max_membership     = TRUE;
static int    fs_EM_stop_criterion_training_LL               = TRUE;
static int    fs_EM_stop_criterion_held_out_LL               = FALSE;
static int    fs_use_initialized_cluster_means_variances_and_priors = FALSE;
static int    fs_max_num_CEM_iterations                      = 100;
static int    fs_write_CEM_intermediate_results              = TRUE;
static int    fs_force_equal_prob_for_CEM_split_and_merge    = TRUE;
static int    fs_crop_feature_dimensions                     = FALSE;
static int    fs_crop_num_feature_dimensions_left            = 0;
static int    fs_crop_num_feature_dimensions_right           = 0;

/* -------------------------------------------------------------------------- */

static int get_independent_GMM_2
(
    int           num_clusters,
    const Matrix* feature_mp,
    Vector**      a_vpp,
    Matrix**      means_mpp,
    Matrix**      var_mpp,
    Matrix**      P_mpp,
    double*       log_likelihood_ptr 
);

static int get_independent_GMM_3
(
    int           num_clusters,
    const Matrix* feature_mp,
    const Vector* held_out_indicator_vp,
    const Vector* initial_a_vp,
    const Matrix* initial_means_mp,
    const Matrix* initial_var_mp,
    Vector**      a_vpp, 
    Matrix**      means_mpp,
    Matrix**      var_mpp,
    Matrix**      P_mpp,
    double*       log_likelihood_ptr,
    double*       held_out_log_likelihood_ptr,
    int*          num_iterations_ptr
);

static int get_discrete_KL_divergence_log_densities
(
    /*const*/ Vector* true_log_distribution_vp,
    /*const*/ Vector* assumed_log_distribution_vp,
    double*       KL_divergence_ptr
);

/* -------------------------------------------------------------------------- */

int set_em_cluster_options(const char* option, const char* value)
{
    char lc_option[ 100 ];
    int  result         = NOT_FOUND;
    int  temp_int_value;
    int  temp_boolean_value;
    double temp_double_value; 
    double   temp_real_value; 
    
    EXTENDED_LC_BUFF_CPY(lc_option, option); 

    if (    (lc_option[ 0 ] == '\0') 
          || match_pattern(lc_option, "cluster-data-perturbation")
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Cluster data perturbation is %.3e.\n", 
                    fs_data_perturbation)); 
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("cluster-data-perturbation = %.3e\n", 
                    fs_data_perturbation)); 
        }
        else
        {
            ERE(ss1snd(value, &temp_double_value));
            fs_data_perturbation = temp_double_value; 
        }
        result = NO_ERROR; 
    }

    if (    (lc_option[ 0 ] == '\0') 
          || match_pattern(lc_option, "cluster-discrete-threshold")
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Cluster threshold for excluding discrete probabilities %.3e.\n", 
                    fs_dis_item_prob_threshold)); 
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("cluster-discrete-threshold = %.3e\n", 
                    fs_dis_item_prob_threshold)); 
        }
        else
        {
            ERE(ss1snd(value, &temp_real_value));
            fs_dis_item_prob_threshold = temp_real_value; 
        }
        result = NO_ERROR; 
    }

    if (    (lc_option[ 0 ] == '\0') 
          || match_pattern(lc_option, "cluster-num-tries-per-cluster-count")
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("cluster-num-tries-per-cluster-count = %d\n", 
                    fs_num_tries_per_cluster_count)); 
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Each number of clusters gets %d EM tries.\n", 
                    fs_num_tries_per_cluster_count)); 
        }
        else
        {
            ERE(ss1pi(value, &temp_int_value));
            if (temp_int_value == 0)
            {
                set_error("Number of tries per cluster count must be at least one."); 
                return ERROR;
            }
            fs_num_tries_per_cluster_count = temp_int_value; 
        }
        result = NO_ERROR; 
    }

    if (    (lc_option[ 0 ] == '\0') 
          || match_pattern(lc_option, "cluster-num-cluster-count-samples")
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("cluster-num-cluster-count-samples = %d\n", 
                    fs_num_cluster_count_samples)); 
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("%d log spaced EM clusters counts are tried.\n", 
                    fs_num_cluster_count_samples)); 
        }
        else
        {
            ERE(ss1pi(value, &temp_int_value));
            if (temp_int_value == 0)
            {
                set_error("Number of EM cluster counts must be at least one."); 
                return ERROR;
            }
            fs_num_cluster_count_samples = temp_int_value; 
        }
        result = NO_ERROR; 
    }

    if (    (lc_option[ 0 ] == '\0') 
          || match_pattern(lc_option, "cluster-max-num-iterations")
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("cluster-max-num-iterations = %d\n", 
                    fs_max_num_iterations)); 
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("A maximum of %d EM iterations are used.\n", 
                    fs_max_num_iterations)); 
        }
        else
        {
            ERE(ss1pi(value, &temp_int_value));
            fs_max_num_iterations = temp_int_value; 
        }
        result = NO_ERROR; 
    }

    if (    (lc_option[ 0 ] == '\0') 
          || match_pattern(lc_option, "cluster-iteration-tolerance")
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("EM iteration tolerance is %.3e.\n", 
                    fs_iteration_tolerance)); 
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("cluster-iteration-tolerance = %.3e\n", fs_iteration_tolerance)); 
        }
        else
        {
            ERE(ss1snd(value, &temp_double_value));
            fs_iteration_tolerance = temp_double_value; 
        }
        result = NO_ERROR; 
    }

    if (    (lc_option[ 0 ] == '\0') 
          || match_pattern(lc_option, "cluster-plot-log-likelihood")
          || match_pattern(lc_option, "cluster-plot-log-likelihood-vs-num-clusters")
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("EM log likelihood vs num clusters %s plotted.\n", 
                    fs_plot_log_likelihood_vs_num_clusters ? "is" : "is not")); 
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("cluster-plot-log-likelihood-vs-num-clusters = %s\n", 
                    fs_plot_log_likelihood_vs_num_clusters ? "t" : "f")); 
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_plot_log_likelihood_vs_num_clusters = temp_boolean_value; 
        }
        result = NO_ERROR; 
    }

    if (    (lc_option[ 0 ] == '\0') 
          || match_pattern(lc_option, "cluster-norm-data")
          || match_pattern(lc_option, "cluster-norm-data-flag")
          || match_pattern(lc_option, "cluster-normalize-data")
          || match_pattern(lc_option, "cluster-normalize-data-flag")
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Cluster data %s normalized.\n", 
                    fs_normalize_data ? "is" : "is not")); 
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("cluster-normalize-data = %s\n", 
                    fs_normalize_data ? "t" : "f")); 
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_normalize_data = temp_boolean_value; 
        }
        result = NO_ERROR; 
    }

    if (    (lc_option[ 0 ] == '\0') 
          || match_pattern(lc_option, "cluster-var-offset")
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Cluster variance offset is %.3e.\n", fs_var_offset)); 
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("cluster-var-offset = %.3e\n", fs_var_offset)); 
        }
        else
        {
            ERE(ss1snd(value, &temp_double_value));
            fs_var_offset = temp_double_value; 
        }
        result = NO_ERROR; 
    }

   if (    (lc_option[ 0 ] == '\0') 
          || match_pattern(lc_option, "cluster-use-unbiased-var-estimate-in-M-step")
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Unbiased estimate for variance %s used in the M step.\n", 
                    fs_use_unbiased_var_estimate_in_M_step ? "is" : "is not"));
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("cluster-use-unbiased-var-estimate-in-M-step = %s\n", 
                    fs_use_unbiased_var_estimate_in_M_step ? "t" : "f")); 
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_use_unbiased_var_estimate_in_M_step = temp_boolean_value; 
        }
        result = NO_ERROR; 
    }

    if (    (lc_option[ 0 ] == '\0') 
          || match_pattern(lc_option, "cluster-held-out-data-fraction")
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Held out data fraction is %.3e.\n", fs_held_out_data_fraction)); 
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("Held out data fraction = %.3e\n", fs_held_out_data_fraction)); 
        }
        else
        {
            ERE(ss1snd(value, &temp_double_value));
            fs_held_out_data_fraction = temp_double_value; 
        }
        result = NO_ERROR; 
    }

    if (    (lc_option[ 0 ] == '\0') 
          || match_pattern(lc_option, "cluster-tie-var")
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Cluster tie variance %s used.\n", 
                    fs_tie_var ? "is" : "is not"));
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("cluster-tie-var = %s\n", 
                    fs_tie_var ? "t" : "f")); 
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_tie_var = temp_boolean_value; 
        }
        result = NO_ERROR; 
    }

    if (    (lc_option[ 0 ] == '\0') 
          || match_pattern(lc_option, "cluster-tie-feature-var")
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Cluster tie feature variance %s used.\n", 
                    fs_tie_feature_var ? "is" : "is not"));
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("cluster-tie-feature-var = %s\n", 
                    fs_tie_feature_var ? "t" : "f")); 
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_tie_feature_var = temp_boolean_value; 
        }
        result = NO_ERROR; 
    }

    if (    (lc_option[ 0 ] == '\0') 
          || match_pattern(lc_option, "cluster-tie-cluster-var")
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Cluster tie cluster variance %s used.\n", 
                    fs_tie_cluster_var ? "is" : "is not"));
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("cluster-tie-cluster-var = %s\n", 
                    fs_tie_cluster_var ? "t" : "f")); 
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_tie_cluster_var = temp_boolean_value; 
        }
        result = NO_ERROR; 
    }

    if (    (lc_option[ 0 ] == '\0') 
          || match_pattern(lc_option, "cluster-model-selection-training-MDL")
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Training data MDL criterion for model selection %s used.\n", 
                    fs_model_selection_training_MDL ? "is" : "is not"));
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("cluster-model-selection-training-MDL = %s\n", 
                    fs_model_selection_training_MDL ? "t" : "f")); 
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_model_selection_training_MDL = temp_boolean_value; 
        }
        result = NO_ERROR; 
    }
    
    if (    (lc_option[ 0 ] == '\0') 
          || match_pattern(lc_option, "cluster-model-selection-held-out-LL")
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Held out data log likelihood criterion for model selection %s used.\n", 
                    fs_model_selection_held_out_LL ? "is" : "is not"));
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("cluster-model-selection-held-out-LL = %s\n", 
                    fs_model_selection_held_out_LL ? "t" : "f")); 
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_model_selection_held_out_LL = temp_boolean_value; 
        }
        result = NO_ERROR; 
    }

    if (    (lc_option[ 0 ] == '\0') 
          || match_pattern(lc_option, "cluster-model-selection-held-out-MDL")
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Held out data MDL criterion for model selection %s used.\n", 
                    fs_model_selection_held_out_MDL ? "is" : "is not"));
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("cluster-model-selection-held-out-MDL = %s\n", 
                    fs_model_selection_held_out_MDL ? "t" : "f")); 
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_model_selection_held_out_MDL = temp_boolean_value; 
        }
        result = NO_ERROR; 
    }
    
    if (    (lc_option[ 0 ] == '\0') 
          || match_pattern(lc_option, "cluster-model-selection-held-out-corr-diff")
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Held out data correlation difference criterion for model selection %s used.\n", 
                    fs_model_selection_held_out_corr_diff ? "is" : "is not"));
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("cluster-model-selection-held-out-corr-diff = %s\n", 
                    fs_model_selection_held_out_corr_diff ? "t" : "f")); 
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_model_selection_held_out_corr_diff = temp_boolean_value; 
        }
        result = NO_ERROR; 
    }

    if (    (lc_option[ 0 ] == '\0') 
          || match_pattern(lc_option, "cluster-model-selection-held-out-max-membership")
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Held out data maximum membership criterion for model selection %s used.\n", 
                    fs_model_selection_held_out_max_membership ? "is" : "is not"));
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("cluster-model-selection-held-out-max-membership = %s\n", 
                    fs_model_selection_held_out_max_membership ? "t" : "f")); 
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_model_selection_held_out_max_membership = temp_boolean_value; 
        }
        result = NO_ERROR; 
    }

    if (    (lc_option[ 0 ] == '\0') 
          || match_pattern(lc_option, "cluster-EM-stop-criterion-training-LL")
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Training data log likelihood %s used as EM stopping criterion.\n", 
                    fs_EM_stop_criterion_training_LL ? "is" : "is not"));
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("cluster-EM-stop-criterion-training-LL = %s\n", 
                    fs_EM_stop_criterion_training_LL ? "t" : "f")); 
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_EM_stop_criterion_training_LL = temp_boolean_value; 
        }
        result = NO_ERROR; 
    }

    if (    (lc_option[ 0 ] == '\0') 
          || match_pattern(lc_option, "cluster-EM-stop-criterion-held-out-LL")
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Held out data log likelihood %s used as EM stopping criterion.\n", 
                    fs_EM_stop_criterion_held_out_LL ? "is" : "is not"));
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("cluster-EM-stop-criterion-held-out-LL = %s\n", 
                    fs_EM_stop_criterion_held_out_LL ? "t" : "f")); 
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_EM_stop_criterion_held_out_LL = temp_boolean_value; 
        }
        result = NO_ERROR; 
    }

    if (    (lc_option[ 0 ] == '\0') 
          || match_pattern(lc_option, "cluster-use-initialized-cluster-means-variances-and-priors")
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Initialized cluster means and variances %s used in the EM algorithm.\n", 
                    fs_use_initialized_cluster_means_variances_and_priors ? "are" : "are not"));
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("cluster-use-initialized-cluster-means-variances-and-priors = %s\n", 
                    fs_use_initialized_cluster_means_variances_and_priors ? "t" : "f")); 
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_use_initialized_cluster_means_variances_and_priors = temp_boolean_value; 
        }
        result = NO_ERROR; 
    }

    if (    (lc_option[ 0 ] == '\0') 
          || match_pattern(lc_option, "cluster-max-num-CEM-iterations")
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("cluster-max-num-CEM-iterations = %d\n", 
                    fs_max_num_CEM_iterations)); 
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("A maximum of %d CEM iterations are used.\n", 
                    fs_max_num_CEM_iterations)); 
        }
        else
        {
            ERE(ss1pi(value, &temp_int_value));
            fs_max_num_CEM_iterations = temp_int_value; 
        }
        result = NO_ERROR;
    }

    if (    (lc_option[ 0 ] == '\0') 
          || match_pattern(lc_option, "cluster-write-CEM-intermediate-results")
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("CEM intermediate results %s written.\n", 
                    fs_write_CEM_intermediate_results ? "are" : "are not"));
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("cluster-write-CEM-intermediate-results = %s\n", 
                    fs_write_CEM_intermediate_results ? "t" : "f")); 
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_write_CEM_intermediate_results = temp_boolean_value; 
        }
        result = NO_ERROR; 
    }

    if (    (lc_option[ 0 ] == '\0') 
          || match_pattern(lc_option, "cluster-force-equal-prob-for-CEM-split-and-merge")
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("The probabilities of choosing a CEM split or a merge operation %s forced to be equal.\n", 
                    fs_force_equal_prob_for_CEM_split_and_merge ? "are" : "are not"));
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("cluster-force-equal-prob-for-CEM-split-and-merge = %s\n", 
                    fs_force_equal_prob_for_CEM_split_and_merge ? "t" : "f")); 
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_force_equal_prob_for_CEM_split_and_merge = temp_boolean_value; 
        }
        result = NO_ERROR; 
    }

    if (    (lc_option[ 0 ] == '\0') 
          || match_pattern(lc_option, "cluster-crop-feature-dimensions")
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Some feature dimensions %s cropped before clustering.\n", 
                    fs_crop_feature_dimensions ? "are" : "are not"));
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("cluster-crop-feature-dimensions = %s\n", 
                    fs_crop_feature_dimensions ? "t" : "f")); 
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_crop_feature_dimensions = temp_boolean_value; 
        }
        result = NO_ERROR; 
    }

    if (    (lc_option[ 0 ] == '\0') 
          || match_pattern(lc_option, "cluster-crop-num-feature-dimensions-left")
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("cluster-crop-num-feature-dimensions-left = %d\n", 
                    fs_crop_num_feature_dimensions_left)); 
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("%d feature dimensions to the left are cropped.\n", 
                    fs_crop_num_feature_dimensions_left)); 
        }
        else
        {
            ERE(ss1pi(value, &temp_int_value));
            fs_crop_num_feature_dimensions_left = temp_int_value; 
        }
        result = NO_ERROR;
    }

    if (    (lc_option[ 0 ] == '\0') 
          || match_pattern(lc_option, "cluster-crop-num-feature-dimensions-right")
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("cluster-crop-num-feature-dimensions-right = %d\n", 
                    fs_crop_num_feature_dimensions_right)); 
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("%d feature dimensions to the right are cropped.\n", 
                    fs_crop_num_feature_dimensions_right)); 
        }
        else
        {
            ERE(ss1pi(value, &temp_int_value));
            fs_crop_num_feature_dimensions_right = temp_int_value; 
        }
        result = NO_ERROR;
    }

    return result; 
}    

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                 get_independent_GMM
 *                               
 * Finds a Gaussian mixture model (GMM).
 *
 * This routine finds a Gaussian mixture model (GMM) for the data on the
 * assumption that the features are independent. The model is fit with EM. Some
 * features are controlled via the set facility. 
 *
 * In particular, it fits:
 * |         p(x) = sum  a-sub-i *  g(u-sub-i, v-sub-i, x)
 * |                 i
 * where a-sub-i is the prior probability for the mixuture compoenent (cluster),
 * u-sub-i is the mean vector for component i, v-sub-i is the variance for the
 * component, and g(u,v,x) is a Gaussian with diagonal covariance (i.e., the
 * features are assumed to be independent, given the cluster). 
 *
 * The argument num_clusters is the number of requested mixture compoenent
 * (clusters), K. 
 *
 * The data matrix feature_mp is an N by M matrix where N is the number of data
 * points, and M is the number of features. 
 *
 * The model parameters are put into *a_vpp, *u_mpp, and *v_mpp. Any of
 * a_vpp, u_mpp, or v_mpp is NULL if that value is not needed.
 *
 * Both u-sub-i and v-sub-i are vectors, and they are put into the i'th row of
 * *u_mpp and *v_ppp, respectively. The matrices are thus K by M. 
 * 
 * If P_mpp, is not NULL, then the soft clustering (cluster membership) for each
 * data point is returned. In that case, *P_mpp will be N by K. 
 *
 * Returns :
 *    If the routine fails (due to storage allocation), then ERROR is returned
 *    with an error message being set. Otherwise NO_ERROR is returned. 
 *
 * Index: clustering, EM, GMM
 *
 * -----------------------------------------------------------------------------
*/

int get_independent_GMM
(
    int           num_clusters,
    const Matrix* feature_mp,
    Vector**      a_vpp,
    Matrix**      u_mpp,
    Matrix**      var_mpp,
    Matrix**      P_mpp
)
{
    int        num_points   = feature_mp->num_rows;
    int        num_features = feature_mp->num_cols;
    int        j;
    double     max_log_likelihood = DBL_HALF_MOST_NEGATIVE;
    Matrix*    u_mp = NULL;
    Matrix*    var_mp = NULL;
    Vector*    a_vp = NULL; 
    Matrix*    P_mp = NULL; 
    Matrix*    perturbed_feature_mp = NULL;
    double     encoding_cost;
    int        result        = NO_ERROR;
    Matrix*        norm_feature_mp = NULL;
    Vector*        feature_mean_vp = NULL;
    Vector*        feature_var_vp  = NULL;
    const double norm_stdev = 1.0; 

    if ((result != ERROR) && (fs_data_perturbation > 10.0 * DBL_EPSILON))
    {
        result = perturb_unary_cluster_data(&perturbed_feature_mp, 
                                            feature_mp, fs_data_perturbation,
                                            (Int_vector*)NULL); 
        feature_mp = perturbed_feature_mp; 
    }

    if (fs_normalize_data)
    {
        result = copy_matrix(&norm_feature_mp, feature_mp);

        if (result != ERROR)
        {
            result = ow_normalize_cluster_data(norm_feature_mp,
                                               (Matrix*)NULL,
                                               &feature_mean_vp, 
                                               &feature_var_vp,
                                               norm_stdev,
                                               (Int_vector*)NULL);
        }

        if (result != ERROR)
        {
            feature_mp = norm_feature_mp; 
        }
    }

    if (fs_num_tries_per_cluster_count <= 1)
    {
        /*
         * Basic simple case. Doing it separately saves memory.
        */
        result = get_independent_GMM_2(num_clusters, 
                                       feature_mp, 
                                       a_vpp, 
                                       u_mpp,
                                       var_mpp, 
                                       P_mpp, 
                                       (double*)NULL); 
    }
    else
    {

        for (j = 0; j < fs_num_tries_per_cluster_count; j++)
        {
            double log_likelihood;

            if (result == ERROR) { NOTE_ERROR(); break; }

            num_clusters = get_independent_GMM_2(num_clusters, 
                                                 feature_mp, 
                                                 (a_vpp == NULL) ? NULL: &a_vp, 
                                                 (u_mpp == NULL) ? NULL : &u_mp, 
                                                 (var_mpp == NULL) ? NULL : &var_mp, 
                                                 (P_mpp == NULL) ? NULL : &P_mp, 
                                                 &log_likelihood); 
            if (num_clusters == ERROR)
            {
                result = ERROR;
                break; 
            }

            verbose_pso(2, "Raw log likelihood with %d clusters is %.5e,\n", 
                        num_clusters, log_likelihood);

            if (fs_tie_var == TRUE)
            {
                encoding_cost = 0.5 * ((num_clusters * (num_features +  1.0) - 1.0) + num_features);
            }
            else if (fs_tie_feature_var == TRUE)
            {
                encoding_cost = 0.5 * ((num_clusters * (num_features +  1.0) - 1.0) + 1.0);
            }
            else
            {
                encoding_cost = 0.5 * (num_clusters * ((2.0 * num_features) +  1.0) - 1.0);
            }
            encoding_cost *= log( (double) num_points*num_features ); 

            log_likelihood -= encoding_cost; 

            verbose_pso(2,
                    "Adjusted log likelihood with %d clusters is %.5e,\n",
                    num_clusters, log_likelihood);

            if (log_likelihood > max_log_likelihood)
            {
                max_log_likelihood = log_likelihood; 

                if ((result != ERROR) && (u_mpp != NULL))
                {
                    result = copy_matrix(u_mpp, u_mp); 
                }

                if ((result != ERROR) && (var_mpp != NULL))
                {
                    result = copy_matrix(var_mpp, var_mp); 
                }
                                                                                                    
                if ((result != ERROR) && (a_vpp != NULL))
                {
                    result = copy_vector(a_vpp, a_vp); 
                }
                                                                                                    
                if ((result != ERROR) && (P_mpp != NULL))
                {
                    result = copy_matrix(P_mpp, P_mp); 
                }
            }

        }
    }

    if ((result != ERROR) && (fs_normalize_data))
    {
        result = ow_un_normalize_cluster_data((u_mpp == NULL) ? NULL : *u_mpp, 
                                              (var_mpp == NULL) ? NULL : *var_mpp, 
                                              feature_mean_vp, 
                                              feature_var_vp, 
                                              norm_stdev); 
    }

    free_matrix(norm_feature_mp);
    free_vector(feature_var_vp); 
    free_vector(feature_mean_vp); 
    free_vector(a_vp);
    free_matrix(u_mp);
    free_matrix(var_mp);
    free_matrix(P_mp);
    free_matrix(perturbed_feature_mp);

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int get_independent_GMM_2
(
    int           num_clusters,
    const Matrix* feature_mp,
    Vector**      a_vpp, 
    Matrix**      means_mpp,
    Matrix**      var_mpp,
    Matrix**      P_mpp,
    double*       log_likelihood_ptr 
)
{
    int           num_points          = feature_mp->num_rows;
    int           num_features        = feature_mp->num_cols;
    int           cluster;
    int           feature;
    double        var_offset = fs_var_offset; 
    int           it, i;
    Vector*       I_vp        = NULL;
    Vector*       a_vp        = NULL;
    Vector*       p_sum_vp    = NULL;
    Matrix*       var_mp        = NULL;
    Matrix*       new_var_mp        = NULL;
    Matrix*       new_u_mp        = NULL;
    Vector*       x_vp        = NULL;
    Vector*       x2_vp        = NULL;
    double        log_likelihood      = DBL_HALF_MOST_NEGATIVE;
    double        prev_log_likelihood = DBL_HALF_MOST_NEGATIVE;
    double        diff;
    Matrix*       u_mp    = NULL;
    Vector*       u_vp    = NULL;
    Vector*       var_vp  = NULL;
    int           result       = NO_ERROR;
#ifdef DONT_USE_GAUSS_ROUTINE
    Vector*       log_sqrt_det_vp  = NULL;
#endif 
    int first_null_cluster_spotted = TRUE; 

    if (P_mpp != NULL)
    {
        ERE(get_target_matrix(P_mpp, num_points, num_clusters));
    }

    if (    (get_target_vector(&I_vp, num_clusters) == ERROR) 
         || (get_target_vector(&a_vp, num_clusters)             == ERROR) 
#ifdef DONT_USE_GAUSS_ROUTINE
         || (get_target_vector(&log_sqrt_det_vp, num_clusters)      == ERROR) 
#endif 
       )
    {
        result = ERROR;
    }

    dbi(num_clusters); 

    for (it = 0; it < fs_max_num_iterations; it++)
    {
        log_likelihood = 0.0;

        if (result == ERROR) { NOTE_ERROR(); break; }

        result = get_zero_vector(&p_sum_vp, num_clusters);
        if (result == ERROR) { NOTE_ERROR(); break; }

        result = get_zero_matrix(&new_u_mp, num_clusters, num_features);
        if (result == ERROR) { NOTE_ERROR(); break; }

        result = get_zero_matrix(&new_var_mp, num_clusters, num_features);
        if (result == ERROR) { NOTE_ERROR(); break; }
                
        for (i = 0; i<num_points; i++)
        {
            if (result == ERROR) { NOTE_ERROR(); break; }

            result = get_matrix_row(&x_vp, feature_mp, i);
            if (result == ERROR) { NOTE_ERROR(); break; } 

            if (it == 0)
            {
                /* E Step init */

                /*
                 * Simple initilaization: One cluster per point, but also make
                 * it so that every cluster gets a tiny bit of weight due to
                 * this point, so that we know that every cluster has some
                 * weight. 
                */
                result = get_random_vector(&I_vp, num_clusters);
                if (result == ERROR) { NOTE_ERROR(); break; }

                result = ow_divide_vector_by_scalar(I_vp, 20.0 * (double)num_points);
                if (result == ERROR) { NOTE_ERROR(); break; }

                cluster = (int)(((double)num_clusters) * kjb_rand());
                if (cluster == num_clusters) cluster--; 
                I_vp->elements[ cluster ] = 1.0; 
                
                result = ow_scale_vector_by_sum(I_vp);
                if (result == ERROR) { NOTE_ERROR(); break; }

                result = ow_add_scalar_to_vector(I_vp, 0.2 * kjb_rand() / num_clusters); 
                if (result == ERROR) { NOTE_ERROR(); break; }

                result = ow_scale_vector_by_sum(I_vp);
                if (result == ERROR) { NOTE_ERROR(); break; }
            }
            else
            {
                /* E Step */

                for (cluster = 0; cluster < num_clusters; cluster++)
                {
                    double log_a = SAFE_LOG(a_vp->elements[ cluster ]);
#ifdef DONT_USE_GAUSS_ROUTINE
                    double* x_ptr = x_vp->elements;
                    double* u_ptr = u_mp->elements[ cluster ];
                    double* v_ptr = var_mp->elements[ cluster ];
                    double temp, dev;
                    double d = 0.0;
#else 

                    if (result == ERROR) { NOTE_ERROR(); break; } 

                    result = get_matrix_row(&u_vp, u_mp, cluster);
                    if (result == ERROR) { NOTE_ERROR(); break; }

                    result = get_matrix_row(&var_vp, var_mp, cluster);
#endif 
                    if (result == ERROR) { NOTE_ERROR(); break; }

#ifdef DONT_USE_GAUSS_ROUTINE
                    for (feature = 0; feature < num_features; feature++)
                    {
                        dev = *u_ptr - *x_ptr;

                        ASSERT_IS_NUMBER_DBL(dev); 
                        ASSERT_IS_FINITE_DBL(dev); 

                        temp = (dev / (*v_ptr)) * dev;

                        d += temp; 

                        ASSERT_IS_NUMBER_DBL(d); 
                        ASSERT_IS_FINITE_DBL(d); 

                        u_ptr++;
                        v_ptr++;
                        x_ptr++;
                    }

                    ASSERT_IS_NUMBER_DBL(d); 
                    ASSERT_IS_FINITE_DBL(d); 

                    I_vp->elements[ cluster ] = log_a - (0.5*d) - log_sqrt_det_vp->elements[ cluster ]; 
#else
                    result = get_log_gaussian_density(x_vp, u_vp, var_vp, 
                                                      &(I_vp->elements[ cluster ]));
                    if (result == ERROR) { NOTE_ERROR(); break; }
                    I_vp->elements[ cluster ] += log_a; 
#endif 

                }

                log_likelihood += ow_exp_scale_by_sum_log_vector(I_vp);
            }

            ASSERT_IS_NEARLY_EQUAL_DBL(sum_vector_elements(I_vp), 1.0, 0.00001);

            if (result == ERROR) { NOTE_ERROR(); break; }

            if (P_mpp != NULL)
            {
                result = put_matrix_row(*P_mpp, I_vp, i);
            }

            /* M-Step */

            result = multiply_vectors(&x2_vp, x_vp, x_vp);
            if (result == ERROR) { NOTE_ERROR(); break; } 

            for (cluster = 0; cluster < num_clusters; cluster++)
            {
                double  p = I_vp->elements[ cluster ];

                if (result == ERROR) { NOTE_ERROR(); break; }

                result = ow_add_scalar_times_vector_to_matrix_row(new_u_mp, 
                                                                  x_vp, p,
                                                                  cluster);
                if (result == ERROR) { NOTE_ERROR(); break; } 

                result = ow_add_scalar_times_vector_to_matrix_row(new_var_mp, 
                                                                  x2_vp, p,
                                                                  cluster);
                if (result == ERROR) { NOTE_ERROR(); break; } 

                p_sum_vp->elements[ cluster ] += p; 
            }
        }

        /* M-Step cleanup */

        if (copy_matrix(&u_mp, new_u_mp) == ERROR)
        {
            db_rv(p_sum_vp); 
            result = ERROR;
            break; 
        }

        for (cluster = 0; cluster < num_clusters; cluster++)
        {
            double s = p_sum_vp->elements[ cluster ];

            if (s > 10.0 * DBL_MIN)
            {
                result = ow_divide_matrix_row_by_scalar(u_mp, s, cluster);
                if (result == ERROR) { NOTE_ERROR(); break; }
            }
            else if (first_null_cluster_spotted)
            {
                first_null_cluster_spotted = FALSE;
                warn_pso("At least one cluster has no members.\n"); 
            }
        }

        result = get_initialized_matrix(&var_mp, num_clusters, num_features, 
                                        DBL_NOT_SET);
        if (result == ERROR) { NOTE_ERROR(); break; }

        for (cluster = 0; cluster < num_clusters; cluster++)
        {
            double s = p_sum_vp->elements[ cluster ];

            for (feature = 0; feature < num_features; feature++)
            {
                double u = u_mp->elements[ cluster ][ feature ];
                double u2 = s * u * u;
                double var = new_var_mp->elements[ cluster ][ feature ] - u2;

                if (var  < 0.0) 
                {
                    /* 
                    // This does happen! The calculation of variance
                    // this way is not numerically stable. 
                    */
                    var = 0.0;
                }

                var_mp->elements[ cluster ][ feature ] = var;
            }

            if (s > 10.0 * DBL_MIN)
            {
                result = ow_divide_matrix_row_by_scalar(var_mp, s, cluster);
                if (result == ERROR) { NOTE_ERROR(); break; }
            }

            verify_matrix(var_mp); 

            /* Moved to later due to the added options of tying variances:
            result = ow_add_scalar_to_matrix_row(var_mp, var_offset, cluster);
            if (result == ERROR) { NOTE_ERROR(); break; }
            */
        }
        
        if (fs_tie_var == TRUE)
        {
            double tie_var = 0.0;
            
            for (feature = 0; feature < num_features; feature++)
            {
                tie_var = 0.0;

                for (cluster = 0; cluster < num_clusters; cluster++)
                {
                    tie_var += var_mp->elements[ cluster ][ feature ];
                }

                tie_var /= num_clusters;

                for (cluster = 0; cluster < num_clusters; cluster++)
                {
                    var_mp->elements[ cluster ][ feature ] = tie_var;
                }
            }
        }
        else if (fs_tie_feature_var == TRUE)
        {
            double tie_feature_var = 0.0;

            for (cluster = 0; cluster < num_clusters; cluster++)
            {
                for (feature = 0; feature < num_features; feature++)
                {
                    tie_feature_var += var_mp->elements[ cluster ][ feature ];
                }
            }

            tie_feature_var /= (num_clusters * num_features);

            for (cluster = 0; cluster < num_clusters; cluster++)
            {
                for (feature = 0; feature < num_features; feature++)
                {
                    var_mp->elements[ cluster ][ feature ] = tie_feature_var;
                }
            }
        }
            
        else if (fs_tie_cluster_var == TRUE)
        {
            for (cluster = 0; cluster < num_clusters; cluster++)
            {
                double tie_cluster_var = 0.0;

                for (feature = 0; feature < num_features; feature++)
                {
                    tie_cluster_var += var_mp->elements[ cluster ][ feature ];
                }

                tie_cluster_var /= num_features;

                for (feature = 0; feature < num_features; feature++)
                {
                    var_mp->elements[ cluster ][ feature ] = tie_cluster_var;
                }
            }
        }
            
        for (cluster = 0; cluster < num_clusters; cluster++)
        {
            result = ow_add_scalar_to_matrix_row(var_mp, var_offset, cluster);
            if (result == ERROR) { NOTE_ERROR(); break; }
        }

#ifdef DONT_USE_GAUSS_ROUTINE
        for (cluster = 0; cluster < num_clusters; cluster++)
        {
            double temp = 0.0;

            if (result == ERROR) { NOTE_ERROR(); break; }

            for (feature = 0; feature < num_features; feature++)
            {
                double var = var_mp->elements[ cluster ][ feature ]; 

                temp += SAFE_LOG(var);
            }

            log_sqrt_det_vp->elements[ cluster ] = temp / 2.0;
        }
        verify_vector(log_sqrt_det_vp);
#endif

        result = scale_vector_by_sum(&a_vp, p_sum_vp); 
        if (result == ERROR) { NOTE_ERROR(); break; } 

        if (it > 0) 
        {
            diff = ABS_OF(log_likelihood - prev_log_likelihood);

            diff *= 2.0;
            diff /= (ABS_OF(log_likelihood) +  ABS_OF(prev_log_likelihood));

            verbose_pso(3, "%-3d: Log likelihood is %12e  |  %10e\n", 
                        it + 1, log_likelihood, diff); 

            if (diff < fs_iteration_tolerance) break; 

            prev_log_likelihood = log_likelihood; 
        }
    }

    if (result != ERROR)
    {
        if (log_likelihood_ptr != NULL)
        {
            *log_likelihood_ptr = log_likelihood;
        }

        if (a_vpp != NULL) 
        {
            result = copy_vector(a_vpp, a_vp);
        }
    }
    
    if ((result != ERROR) && (var_mpp != NULL))
    {
        result = copy_matrix(var_mpp, var_mp);
    }

    if ((result != ERROR) && (means_mpp != NULL))
    {
        result = copy_matrix(means_mpp, u_mp);
    }

    free_vector(I_vp); 
    free_vector(p_sum_vp);
    free_vector(a_vp);
    free_matrix(new_u_mp); 
    free_matrix(u_mp); 
    free_matrix(new_var_mp); 
    free_matrix(var_mp); 
#ifdef DONT_USE_GAUSS_ROUTINE
    free_vector(log_sqrt_det_vp);
#endif 

    free_vector(x2_vp); 
    free_vector(x_vp); 
    free_vector(var_vp); 
    free_vector(u_vp); 

    if (result == ERROR) return ERROR; else return num_clusters;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int get_independent_GMM_3
(
    int           num_clusters,
    const Matrix* feature_mp,
    const Vector* held_out_indicator_vp,
    const Vector* initial_a_vp,
    const Matrix* initial_means_mp,
    const Matrix* initial_var_mp,
    Vector**      a_vpp, 
    Matrix**      means_mpp,
    Matrix**      var_mpp,
    Matrix**      P_mpp,
    double*       log_likelihood_ptr,
    double*       held_out_log_likelihood_ptr,
    int*          num_iterations_ptr
)
{
    int           num_points          = feature_mp->num_rows;
    int           num_features        = feature_mp->num_cols;
    int           cluster;
    int           feature;
    double        var_offset = fs_var_offset; 
    int           it, i;
    Vector*       I_vp          = NULL;
    Vector*       a_vp        = NULL;
    Vector*       p_sum_vp    = NULL;
    Matrix*       var_mp        = NULL;
    Matrix*       new_var_mp        = NULL;
    Matrix*       new_u_mp        = NULL;
    Vector*       x_vp        = NULL;
    Vector*       x2_vp        = NULL;
    double        log_likelihood      = DBL_HALF_MOST_NEGATIVE;
    double        prev_log_likelihood = DBL_HALF_MOST_NEGATIVE;
    double        held_out_log_likelihood      = DBL_HALF_MOST_NEGATIVE;
    double        prev_held_out_log_likelihood = DBL_HALF_MOST_NEGATIVE;
    double        diff, held_out_diff;
    Matrix*       u_mp    = NULL;
    Vector*       u_vp    = NULL;
    Vector*       var_vp  = NULL;
    int           result       = NO_ERROR;
#ifdef DONT_USE_GAUSS_ROUTINE
    Vector*       log_sqrt_det_vp  = NULL;
#endif 
    Vector*       p_square_sum_vp = NULL;

    if (P_mpp != NULL)
    {
        ERE(get_target_matrix(P_mpp, num_points, num_clusters));
    }

    if (held_out_indicator_vp != NULL)
    {
        if (held_out_indicator_vp->length != num_points)
        {
            result = ERROR;
        }
    }
            
    if (    (get_target_vector(&I_vp, num_clusters) == ERROR) 
         || (get_target_vector(&a_vp, num_clusters) == ERROR) 
#ifdef DONT_USE_GAUSS_ROUTINE
         || (get_target_vector(&log_sqrt_det_vp, num_clusters) == ERROR) 
#endif 
       )
    {
        result = ERROR;
    }

    dbi(num_clusters); 

    for (it = 0; it < fs_max_num_iterations; it++)
    {
        log_likelihood = 0.0;

        held_out_log_likelihood = 0.0;

        if (result == ERROR) { NOTE_ERROR(); break; }

        result = get_zero_vector(&p_sum_vp, num_clusters);
        if (result == ERROR) { NOTE_ERROR(); break; }

        result = get_zero_vector(&p_square_sum_vp, num_clusters);
        if (result == ERROR) { NOTE_ERROR(); break; }

        result = get_zero_matrix(&new_u_mp, num_clusters, num_features);
        if (result == ERROR) { NOTE_ERROR(); break; }

        result = get_zero_matrix(&new_var_mp, num_clusters, num_features);
        if (result == ERROR) { NOTE_ERROR(); break; }
                
        if ( (it == 0) && (fs_use_initialized_cluster_means_variances_and_priors == TRUE) )
        {
            if (initial_means_mp != NULL)
            {   
                ERE(get_target_matrix(&u_mp, num_clusters, num_features));
                for (i = 0; i < num_clusters; i++)
                {
                    ERE(copy_matrix_row(u_mp, i, initial_means_mp, i));
                }
            }
            else
            {
                 set_bug("Initial cluster means matrix is NULL.\n");
                 return ERROR;
                 
                /*
                warn_pso("Initial cluster means matrix is NULL: Using random means.\n");
                ERE(get_random_matrix(&u_mp, num_clusters, num_features));
                */
            }

            if (initial_var_mp != NULL)
            {
                ERE(get_target_matrix(&var_mp, num_clusters, num_features));
                for (i = 0; i < num_clusters; i++)
                {
                    ERE(copy_matrix_row(var_mp, i, initial_var_mp, i));
                }
            }
            else
            {
                set_bug("Initial cluster variances matrix is NULL.\n");
                return ERROR;
                
                /*
                warn_pso("Initial cluster variances matrix is NULL: Using random variances.\n");
                ERE(get_random_matrix(&var_mp, num_clusters, num_features));
                */
            }
            for (i = 0; i < num_clusters; i++)
            {
                ERE(ow_add_scalar_to_matrix_row(var_mp, var_offset, i));
            }

            if (initial_a_vp != NULL)
            {
                ERE(get_target_vector(&a_vp, num_clusters));
                for (i = 0; i < num_clusters; i++)
                {
                    a_vp->elements[i] = initial_a_vp->elements[i];
                }
                ERE(ow_normalize_vector(a_vp, NORMALIZE_BY_SUM));
            }
            else
            {
                set_bug("Initial cluster priors vector is NULL.\n");
                return ERROR;
                
                /*
                warn_pso("Initial cluster priors vector is NULL: Using random priors.\n");
                for (i = 0; i < num_clusters; i++)
                {
                    a_vp->elements[i] = kjb_rand();
                }
                ERE(ow_add_scalar_to_vector(a_vp, 0.2 * kjb_rand() / num_clusters)); 
                ERE(ow_normalize_vector(a_vp, NORMALIZE_BY_SUM));
                */
            }

#ifdef DONT_USE_GAUSS_ROUTINE
            for (cluster = 0; cluster < num_clusters; cluster++)
            {
                double temp = 0.0;

                if (result == ERROR) { NOTE_ERROR(); break; }

                for (feature = 0; feature < num_features; feature++)
                {
                    double var = var_mp->elements[ cluster ][ feature ]; 

                    temp += SAFE_LOG(var);
                }

                log_sqrt_det_vp->elements[ cluster ] = temp / 2.0;
            }
            verify_vector(log_sqrt_det_vp);
#endif 
        }
        
        for (i = 0; i<num_points; i++)
        {
            if (result == ERROR) { NOTE_ERROR(); break; }

            result = get_matrix_row(&x_vp, feature_mp, i);
            if (result == ERROR) { NOTE_ERROR(); break; } 

            if ( (it == 0) && (fs_use_initialized_cluster_means_variances_and_priors != TRUE) )
            {
                /* E Step init */

                result = get_target_vector(&I_vp, num_clusters);
                if (result == ERROR) { NOTE_ERROR(); break; }
                
                for (cluster = 0; cluster < num_clusters; cluster++)
                {
                    double r = kjb_rand();
                    double p = pow(r, 5.0); /* Perhaps should be an option. */

                    if (result == ERROR) { NOTE_ERROR(); break; }

                    I_vp->elements[ cluster ] = p;

                    if (kjb_rand() < 1.0 / num_clusters) 
                    {
                        I_vp->elements[ cluster ] += 0.5; 
                    }
                }

                result = ow_scale_vector_by_sum(I_vp);
                if (result == ERROR) { NOTE_ERROR(); break; }

                result = ow_add_scalar_to_vector(I_vp, 0.2 * kjb_rand() / num_clusters); 
                if (result == ERROR) { NOTE_ERROR(); break; }

                result = ow_scale_vector_by_sum(I_vp);
                if (result == ERROR) { NOTE_ERROR(); break; }
            }
            
            if ( (it > 0) || (fs_use_initialized_cluster_means_variances_and_priors == TRUE) )
            {
                /* E Step */

                for (cluster = 0; cluster < num_clusters; cluster++)
                {
                    double log_a = SAFE_LOG(a_vp->elements[ cluster ]);
#ifdef DONT_USE_GAUSS_ROUTINE
                    double* x_ptr = x_vp->elements;
                    double* u_ptr = u_mp->elements[ cluster ];
                    double* v_ptr = var_mp->elements[ cluster ];
                    double temp, dev;
                    double d = 0.0;
#else 

                    if (result == ERROR) { NOTE_ERROR(); break; } 

                    result = get_matrix_row(&u_vp, u_mp, cluster);
                    if (result == ERROR) { NOTE_ERROR(); break; }

                    result = get_matrix_row(&var_vp, var_mp, cluster);
#endif 
                    if (result == ERROR) { NOTE_ERROR(); break; }

#ifdef DONT_USE_GAUSS_ROUTINE
                    for (feature = 0; feature < num_features; feature++)
                    {
                        dev = *u_ptr - *x_ptr;

                        ASSERT_IS_NUMBER_DBL(dev); 
                        ASSERT_IS_FINITE_DBL(dev); 

                        temp = (dev / (*v_ptr)) * dev;

                        d += temp; 

                        ASSERT_IS_NUMBER_DBL(d); 
                        ASSERT_IS_FINITE_DBL(d); 

                        u_ptr++;
                        v_ptr++;
                        x_ptr++;
                    }

                    ASSERT_IS_NUMBER_DBL(d); 
                    ASSERT_IS_FINITE_DBL(d); 

                    I_vp->elements[ cluster ] = log_a - (0.5*d) - log_sqrt_det_vp->elements[ cluster ]; 
#else
                    result = get_log_gaussian_density(x_vp, u_vp, var_vp, 
                                                      &(I_vp->elements[ cluster ]));
                    if (result == ERROR) { NOTE_ERROR(); break; }
                    I_vp->elements[ cluster ] += log_a; 
#endif 

                }

                if ( (held_out_indicator_vp != NULL) && (held_out_indicator_vp->elements[i] == 1) )
                {
                    held_out_log_likelihood += ow_exp_scale_by_sum_log_vector(I_vp);
                }
                else
                {
                    log_likelihood += ow_exp_scale_by_sum_log_vector(I_vp);
                }
            }

            ASSERT_IS_NEARLY_EQUAL_DBL(sum_vector_elements(I_vp), 1.0, 0.00001);

            if (result == ERROR) { NOTE_ERROR(); break; }

            if (P_mpp != NULL)
            {
                result = put_matrix_row(*P_mpp, I_vp, i);
            }

            /* M-Step */
            
            if ( (held_out_indicator_vp != NULL) && (held_out_indicator_vp->elements[i] == 1) )
            {
                /* Do not do the M-step for held out points */
            }
            else
            {
                result = multiply_vectors(&x2_vp, x_vp, x_vp);
                if (result == ERROR) { NOTE_ERROR(); break; } 

                for (cluster = 0; cluster < num_clusters; cluster++)
                {
                    double  p = I_vp->elements[ cluster ];

                    if (result == ERROR) { NOTE_ERROR(); break; }

                    result = ow_add_scalar_times_vector_to_matrix_row(new_u_mp, 
                                                                      x_vp, p,
                                                                      cluster);
                    if (result == ERROR) { NOTE_ERROR(); break; } 
    
                    result = ow_add_scalar_times_vector_to_matrix_row(new_var_mp, 
                                                                      x2_vp, p,
                                                                      cluster);
                    if (result == ERROR) { NOTE_ERROR(); break; } 

                    p_sum_vp->elements[ cluster ] += p; 

                    p_square_sum_vp->elements[ cluster ] += (p * p);
                }
            }
        }

        /* M-Step cleanup */

        if (divide_matrix_by_col_vector(&u_mp, new_u_mp, p_sum_vp) == ERROR)
        {
            db_rv(p_sum_vp); 
            result = ERROR;
            break; 
        }

        result = get_initialized_matrix(&var_mp, num_clusters, num_features, 
                                        DBL_NOT_SET);
        if (result == ERROR) { NOTE_ERROR(); break; }

        for (cluster = 0; cluster < num_clusters; cluster++)
        {
            double s = p_sum_vp->elements[ cluster ];

            for (feature = 0; feature < num_features; feature++)
            {
                double u = u_mp->elements[ cluster ][ feature ];
                double u2 = s * u * u;
                double var = new_var_mp->elements[ cluster ][ feature ] - u2;

                if (var  < 0.0) 
                {
                    /* 
                    // This does happen! The calculation of variance
                    // this way is not numerically stable. 
                    */
                    var = 0.0;
                }

                var_mp->elements[ cluster ][ feature ] = var;
            }

            result = ow_divide_matrix_row_by_scalar(var_mp, s, cluster);
            if (result == ERROR) { NOTE_ERROR(); break; }

            if (fs_use_unbiased_var_estimate_in_M_step == TRUE)
            {
                double p_square_sum      = p_square_sum_vp->elements[ cluster ];
                double norm_p_square_sum = p_square_sum / (s * s);
                double norm_factor       = 1.0 / (1.0 - norm_p_square_sum);

                result = ow_multiply_matrix_row_by_scalar(var_mp, norm_factor, cluster);
                if (result == ERROR) { NOTE_ERROR(); break; }
            }
            
            verify_matrix(var_mp); 

            /* Moved to later due to the added options of tying variances:
            result = ow_add_scalar_to_matrix_row(var_mp, var_offset, cluster);
            if (result == ERROR) { NOTE_ERROR(); break; }
            */
        }

        if (fs_tie_var == TRUE)
        {
            double tie_var = 0.0;
            
            for (feature = 0; feature < num_features; feature++)
            {
                tie_var = 0.0;

                for (cluster = 0; cluster < num_clusters; cluster++)
                {
                    tie_var = tie_var + var_mp->elements[ cluster ][ feature ];
                }

                tie_var = tie_var/num_clusters;

                for (cluster = 0; cluster < num_clusters; cluster++)
                {
                    var_mp->elements[ cluster ][ feature ] = tie_var;
                }
            }
        }
        else if (fs_tie_feature_var == TRUE)
        {
            double tie_feature_var = 0.0;

            for (cluster = 0; cluster < num_clusters; cluster++)
            {
                for (feature = 0; feature < num_features; feature++)
                {
                    tie_feature_var = tie_feature_var + var_mp->elements[ cluster ][ feature ];
                }
            }

            tie_feature_var = tie_feature_var/(num_clusters * num_features);

            for (cluster = 0; cluster < num_clusters; cluster++)
            {
                for (feature = 0; feature < num_features; feature++)
                {
                    var_mp->elements[ cluster ][ feature ] = tie_feature_var;
                }
            }
        }
            
        for (cluster = 0; cluster < num_clusters; cluster++)
        {
            result = ow_add_scalar_to_matrix_row(var_mp, var_offset, cluster);
            if (result == ERROR) { NOTE_ERROR(); break; }
        }
        
#ifdef DONT_USE_GAUSS_ROUTINE
        for (cluster = 0; cluster < num_clusters; cluster++)
        {
            double temp = 0.0;

            if (result == ERROR) { NOTE_ERROR(); break; }

            for (feature = 0; feature < num_features; feature++)
            {
                double var = var_mp->elements[ cluster ][ feature ]; 

                temp += SAFE_LOG(var);
            }

            log_sqrt_det_vp->elements[ cluster ] = temp / 2.0;
        }
        verify_vector(log_sqrt_det_vp);
#endif 

        result = scale_vector_by_sum(&a_vp, p_sum_vp); 
        if (result == ERROR) { NOTE_ERROR(); break; } 

        if (it > 0) 
        {
            diff = ABS_OF(log_likelihood - prev_log_likelihood);

            diff *= 2.0;
            diff /= (ABS_OF(log_likelihood) +  ABS_OF(prev_log_likelihood));

            verbose_pso(3, "%-3d: Log likelihood is %12e  |  %10e\n", 
                        it + 1, log_likelihood, diff); 

            held_out_diff = ABS_OF(held_out_log_likelihood - prev_held_out_log_likelihood);

            held_out_diff *= 2.0;
            held_out_diff /= (ABS_OF(held_out_log_likelihood) + ABS_OF(prev_held_out_log_likelihood));
            
            verbose_pso(3, "%-3d: Held out log likelihood is %12e  |  %10e\n",
                        it + 1, held_out_log_likelihood, held_out_diff); 

            if (fs_EM_stop_criterion_training_LL == TRUE)
            {
                if (diff < fs_iteration_tolerance) break; 
            }
            else if (fs_EM_stop_criterion_held_out_LL == TRUE)
            {
                if (held_out_diff < fs_iteration_tolerance) break;
            }

            prev_log_likelihood = log_likelihood;

            prev_held_out_log_likelihood = held_out_log_likelihood;
        }
    }

    if (result != ERROR)
    {
        if (log_likelihood_ptr != NULL)
        {
            *log_likelihood_ptr = log_likelihood;
        }

        if (held_out_log_likelihood_ptr != NULL)
        {
            *held_out_log_likelihood_ptr = held_out_log_likelihood;
        }

        if (num_iterations_ptr != NULL)
        {
            *num_iterations_ptr = it;
        }

        if (a_vpp != NULL) 
        {
            result = copy_vector(a_vpp, a_vp);
        }
    }
    
    if ((result != ERROR) && (var_mpp != NULL))
    {
        result = copy_matrix(var_mpp, var_mp);
    }

    if ((result != ERROR) && (means_mpp != NULL))
    {
        result = copy_matrix(means_mpp, u_mp);
    }

    free_vector(I_vp); 
    free_vector(p_sum_vp);
    free_vector(a_vp);
    free_matrix(new_u_mp); 
    free_matrix(u_mp); 
    free_matrix(new_var_mp); 
    free_matrix(var_mp); 
#ifdef DONT_USE_GAUSS_ROUTINE
    free_vector(log_sqrt_det_vp);
#endif 

    free_vector(x2_vp); 
    free_vector(x_vp); 
    free_vector(var_vp); 
    free_vector(u_vp); 

     free_vector(p_square_sum_vp);

    if (result == ERROR) 
    {
        return ERROR; 
    }
    else 
    {
        return num_clusters;
    }
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                 get_independent_GMM_using_CEM
 *                               
 * Finds a Gaussian mixture model (GMM) by automatically choosing an optimal
 * number of clusters.
 *
 * This routine finds a Gaussian mixture model (GMM) for the data on the
 * assumption that the features are independent. It starts with an initial
 * number of clusters and possibly proceeds through a series of split-merge
 * operations on the clusters to choose an optimal number of clusters. The
 * optimizing criterion for choosing the final number of clusters is based on a
 * trade-off between the log-likelihood and an information theoretic criterion
 * as described in the following paper:
 * 
 * Baibo Zhang, Changshui Zhang, Xing Yi, "Competitive EM Algorithm for Finite
 * Mixture Models", Pattern Recognition, PR(37), No. 1, Jan. 2004, pp. 131-144.
 * 
 * Some features of this routine are controlled via the set facility. 
 *
 * In particular, it fits:
 * |         p(x) = sum  a-sub-i *  g(u-sub-i, v-sub-i, x)
 * |                 i
 * where a-sub-i is the prior probability for the mixuture component (cluster),
 * u-sub-i is the mean vector for component i, v-sub-i is the variance for the
 * component, and g(u,v,x) is a Gaussian with diagonal covariance (i.e., the
 * features are assumed to be independent, given the cluster). 
 *
 * The argument initial_num_clusters is the initial number of mixture components
 * (clusters). 
 *
 * The data matrix feature_mp is an N by M matrix where N is the number of data
 * points, and M is the number of features. 
 *
 * The means, variances and priors of the initial clusters can be specified
 * through the arguments initial_u_mp, initial_var_mp, and initial_a_vp
 * respectively. If they are all NULL, then a random initialization method is used
 * for the clusters.
 *
 * The model parameters are put into *a_vpp, *u_mpp, and *v_mpp. Any of
 * a_vpp, u_mpp, or v_mpp is NULL if that value is not needed.
 *
 * Both u-sub-i and v-sub-i are vectors, and they are put into the i'th row of
 * *u_mpp and *v_mpp, respectively. The matrices are thus K by M, where K is the
 * final number of clusters.
 * 
 * If P_mpp, is not NULL, then the soft clustering (cluster membership) for each
 * data point is returned. In that case, *P_mpp will be N by K. 
 *
 * The parameters gamma and beta control the cluster split-merge operations.
 * These need to be determined experimentally. However, the following
 * observations might help as a rough guide in the selection of values for these:
 * 
 * The higher the value of gamma, the greater is the frequency of split-merge
 * operations and hence greater is the ability to jump in the parameter space (see
 * reference). A value that is of the order of 0.1L or L, where L is the
 * adjusted log-likelihood, has been observed to give reasonable results. The
 * order of L can be determined from the logs printed out by the routine.
 *
 * The higher the value of beta, the greater is the chance of merge operations
 * being chosen relative to split operations (see reference) while proceeding
 * through the split-merge sequence. A value in the range of 0.1R - 10R, where,
 * R = min(KL_divergence_merge^2, KL_divergence_split^2), has been observed to give
 * reasonable results. R has to be determined experimentally by observing the KL
 * divergence values printed out by the routine.
 *
 * Warning:
 *    This routine might lead to memory leaks if fails to execute completely.
 *
 * Returns:
 *    If the routine fails, then ERROR is returned with an error message being 
 *    set. Otherwise NO_ERROR is returned. 
 *
 * Index: clustering, EM, GMM
 *
 * -----------------------------------------------------------------------------
*/

int get_independent_GMM_using_CEM
(
    int             initial_num_clusters,
    const Matrix*   feature_mp,
    const Vector*   initial_a_vp,
    const Matrix*   initial_u_mp,
    const Matrix*   initial_var_mp,
    Vector**        a_vpp,
    Matrix**        u_mpp,
    Matrix**        var_mpp,
    Matrix**        P_mpp,
    double          beta,
    double          gamma
)
{
    int        num_points   = feature_mp->num_rows;
    int        num_features = feature_mp->num_cols;
    int        i, j, k, l, m, n, cluster, feature, point;
    int        num_clusters, new_num_clusters;
    
    Vector*    a_vp = NULL; 
    Matrix*    u_mp = NULL;
    Matrix*    var_mp = NULL;
    Matrix*    local_P_mp = NULL; 
    Matrix**   local_P_mpp = (P_mpp == NULL) ? NULL : &local_P_mp; 
    Matrix*    perturbed_feature_mp = NULL;
    
    Matrix*    norm_feature_mp = NULL; 
    Vector*    feature_mean_vp = NULL;
    Vector*    feature_var_vp = NULL;
    
    const double norm_stdev = 1.0;

    int    num_E_M_iterations, max_num_E_M_iterations;
    int    use_initialization_E_M;
    double log_likelihood, held_out_log_likelihood, adjusted_log_likelihood;
    double new_log_likelihood, new_held_out_log_likelihood, adjusted_new_log_likelihood;
    double max_adjusted_log_likelihood = DBL_HALF_MOST_NEGATIVE;
    
    int    num_possible_split_merge_operations;
    double K_L_divergence;
    
    Int_matrix* split_merge_cluster_indices_mp         = NULL;
    Vector*     split_merge_probabilities_vp           = NULL;
    Vector*     split_merge_cumulative_distribution_vp = NULL;
    Vector*     weighted_empirical_distribution_vp     = NULL;
    Vector*     weighted_log_empirical_distribution_vp = NULL;
    Vector*     log_sqrt_det_vp                        = NULL;
    Vector*     x_vp                                   = NULL;
    Matrix*     log_cluster_distributions_mp           = NULL;
    Vector*     cluster_distribution_vp                = NULL;
    Vector*     log_cluster_distribution_vp            = NULL;

    double temp;
    
    double split_merge_a, split_merge_log_sqrt_det;
    
    Vector*     split_merge_u_vp   = NULL;
    Vector*     split_merge_var_vp = NULL;

    Vector*     u_1_vp   = NULL;
    Vector*     u_2_vp   = NULL;
    Vector*     var_1_vp = NULL;
    Vector*     var_2_vp = NULL;

    Vector*     merge_log_cluster_distribution_vp = NULL;
    Matrix*     merge_P_mp                        = NULL;
    Vector*     temp_vp                           = NULL;
    
    double random_number;

    int split_cluster_index, max_var_index;
    int merge_cluster_index_1, merge_cluster_index_2;
    double max_var;
    
    Vector*     local_initial_a_vp   = NULL;
    Matrix*     local_initial_u_mp   = NULL;
    Matrix*     local_initial_var_mp = NULL;
    
    double acceptance_prob;

    int    num_parameters_per_cluster;
    int    num_clusters_annihilate;
    double total_mass_annihilate;
 
    Matrix_vector* intermediate_a_mvp   = NULL;
    Matrix_vector* intermediate_u_mvp   = NULL;
    Matrix_vector* intermediate_var_mvp = NULL;

    Matrix* intermediate_a_mp   = NULL;
    Matrix* intermediate_u_mp   = NULL;
    Matrix* intermediate_var_mp = NULL;
    
    int intermediate_result_index;

    double total_split_KL_divergence, total_merge_inverse_KL_divergence;
    
    if ( fs_data_perturbation > 10.0 * DBL_EPSILON )
    {
        ERE(perturb_unary_cluster_data(&perturbed_feature_mp, feature_mp, fs_data_perturbation, (Int_vector*)NULL)); 
        feature_mp = perturbed_feature_mp; 
    }

    /* Normalization to unit variance in all dimensions. */
    if (fs_normalize_data)
    {
        ERE(copy_matrix(&norm_feature_mp, feature_mp));
        
        ERE(ow_normalize_cluster_data(norm_feature_mp,
                                      (Matrix*)NULL,
                                      &feature_mean_vp, 
                                      &feature_var_vp,
                                      norm_stdev,
                                      (Int_vector*)NULL));
        
        db_rv(feature_var_vp); 
        
        feature_mp = norm_feature_mp; 
    }
    
    num_parameters_per_cluster = num_features + num_features + 1;
    
    /* Compute a first local maxima in the parameter space starting from initial
     * parameter values. */
    num_clusters = initial_num_clusters;

    ERE(copy_vector(&local_initial_a_vp, initial_a_vp));
    ERE(copy_matrix(&local_initial_u_mp, initial_u_mp));
    ERE(copy_matrix(&local_initial_var_mp, initial_var_mp)); 
    pso("First run of EM using %d clusters.\n", num_clusters);

    max_num_E_M_iterations = fs_max_num_iterations;
    use_initialization_E_M = fs_use_initialized_cluster_means_variances_and_priors; 

    /* Keep track of intermediate results. */
    if (fs_write_CEM_intermediate_results == TRUE)
    {
        /* Allocate memory to store intermediate CEM results:
         * Assume that 5 times the total number of CEM iterations will be
         * the maximum number of intermediate results that will be stored.
         * Since the intermediate results being stored include configurations:
         * 1. before and after each EM iteration.
         * 2. before and after every cluster annihilation step.
         * at feasible points in time, the total number of such results cannot
         * be predicted beforehand.
         * 
         * Therefore the above assumption of 5 times the total number of CEM
         * iterations might lead to accessing unallocated memory!!!!!!!!!!!!!!!
         */
        
        ERE(get_target_matrix_vector(&intermediate_a_mvp, 5*fs_max_num_CEM_iterations));
        ERE(get_target_matrix_vector(&intermediate_u_mvp, 5*fs_max_num_CEM_iterations));
        ERE(get_target_matrix_vector(&intermediate_var_mvp, 5*fs_max_num_CEM_iterations));
    }
    
    intermediate_result_index = 0;
    
    /* Keep track of intermediate results. */
    if (fs_write_CEM_intermediate_results == TRUE)
    {
        if ( (local_initial_a_vp != NULL) && (local_initial_u_mp != NULL) && (local_initial_var_mp != NULL) )
        {
            ERE(get_target_matrix(&intermediate_a_mvp->elements[intermediate_result_index], num_clusters, 1));
            ERE(get_target_matrix(&intermediate_u_mvp->elements[intermediate_result_index], num_clusters, num_features));
            ERE(get_target_matrix(&intermediate_var_mvp->elements[intermediate_result_index], num_clusters, num_features));

            intermediate_a_mp = intermediate_a_mvp->elements[intermediate_result_index];
            intermediate_u_mp = intermediate_u_mvp->elements[intermediate_result_index];
            intermediate_var_mp = intermediate_var_mvp->elements[intermediate_result_index];

            for (cluster = 0; cluster < num_clusters; cluster++)
            {
                intermediate_a_mp->elements[cluster][0] = local_initial_a_vp->elements[cluster];
                ERE(copy_matrix_row(intermediate_u_mp, cluster, local_initial_u_mp, cluster));
                ERE(copy_matrix_row(intermediate_var_mp, cluster, local_initial_var_mp, cluster));
            }

            intermediate_result_index = intermediate_result_index + 1; 
        }
    }
    
    do
    {
        ERE(num_clusters = get_independent_GMM_3(num_clusters, 
                                                 feature_mp,
                                                 (Vector*) NULL,
                                                 local_initial_a_vp, 
                                                 local_initial_u_mp,
                                                 local_initial_var_mp,
                                                 &a_vp, &u_mp, &var_mp,  
                                                 local_P_mpp, 
                                                 &log_likelihood,
                                                 &held_out_log_likelihood,
                                                 &num_E_M_iterations));

        fs_max_num_iterations = max_num_E_M_iterations;
        fs_use_initialized_cluster_means_variances_and_priors = use_initialization_E_M;

        /* Keep track of intermediate results. */
        if (fs_write_CEM_intermediate_results == TRUE)
        {
            if ( (a_vp != NULL) && (u_mp != NULL) && (var_mp != NULL) )
            {
                ERE(get_target_matrix(&intermediate_a_mvp->elements[intermediate_result_index], num_clusters, 1));
                ERE(get_target_matrix(&intermediate_u_mvp->elements[intermediate_result_index], num_clusters, num_features));
                ERE(get_target_matrix(&intermediate_var_mvp->elements[intermediate_result_index], num_clusters, num_features));

                intermediate_a_mp = intermediate_a_mvp->elements[intermediate_result_index];
                intermediate_u_mp = intermediate_u_mvp->elements[intermediate_result_index];
                intermediate_var_mp = intermediate_var_mvp->elements[intermediate_result_index];

                for (cluster = 0; cluster < num_clusters; cluster++)
                {
                    intermediate_a_mp->elements[cluster][0] = a_vp->elements[cluster];
                    ERE(copy_matrix_row(intermediate_u_mp, cluster, u_mp, cluster));
                    ERE(copy_matrix_row(intermediate_var_mp, cluster, var_mp, cluster));
                }

                intermediate_result_index = intermediate_result_index + 1; 
            }
        }

        /* Cluster annihilation. */
        num_clusters_annihilate = 0;
        total_mass_annihilate   = 0.0;
        for (cluster = 0; cluster < num_clusters; cluster++)
        {
            if ( (a_vp->elements[cluster] * num_points) < num_parameters_per_cluster )
            {
                ERE(remove_matrix_row((Vector**) NULL, u_mp, (cluster - num_clusters_annihilate)));
                ERE(remove_matrix_row((Vector**) NULL, var_mp, (cluster - num_clusters_annihilate)));
                
                pso("Annihilating cluster %d with posterior %f.\n", cluster, a_vp->elements[cluster]);
                num_clusters_annihilate = num_clusters_annihilate + 1;
                total_mass_annihilate = total_mass_annihilate + a_vp->elements[cluster];
            }
        }

        ERE(get_target_vector(&temp_vp, (num_clusters - num_clusters_annihilate)));
        for (cluster = 0, m = 0; cluster < num_clusters; cluster++)
        {
            if ( (a_vp->elements[cluster] * num_points) >= num_parameters_per_cluster )
            {
                temp_vp->elements[m] = a_vp->elements[cluster] + (total_mass_annihilate/(num_clusters - num_clusters_annihilate));
                m = m+1;
            }
        }
        ERE(scale_vector_by_sum(&a_vp, temp_vp));

        num_clusters = num_clusters - num_clusters_annihilate;

        if (num_clusters_annihilate > 0)
        {
            ERE(copy_vector(&local_initial_a_vp, a_vp));
            ERE(copy_matrix(&local_initial_u_mp, u_mp));
            ERE(copy_matrix(&local_initial_var_mp, var_mp));
        
            /* Keep track of intermediate results. */
            if (fs_write_CEM_intermediate_results == TRUE)
            {
                if ( (a_vp != NULL) && (u_mp != NULL) && (var_mp != NULL) )
                {
                    ERE(get_target_matrix(&intermediate_a_mvp->elements[intermediate_result_index], num_clusters, 1));
                    ERE(get_target_matrix(&intermediate_u_mvp->elements[intermediate_result_index], num_clusters, num_features));
                    ERE(get_target_matrix(&intermediate_var_mvp->elements[intermediate_result_index], num_clusters, num_features));

                    intermediate_a_mp = intermediate_a_mvp->elements[intermediate_result_index];
                    intermediate_u_mp = intermediate_u_mvp->elements[intermediate_result_index];
                    intermediate_var_mp = intermediate_var_mvp->elements[intermediate_result_index];

                    for (cluster = 0; cluster < num_clusters; cluster++)
                    {
                        intermediate_a_mp->elements[cluster][0] = a_vp->elements[cluster];
                        ERE(copy_matrix_row(intermediate_u_mp, cluster, u_mp, cluster));
                        ERE(copy_matrix_row(intermediate_var_mp, cluster, var_mp, cluster));
                    }

                    intermediate_result_index = intermediate_result_index + 1; 
                }
            }

            /* This is necessary because the first run of EM might not have used
               initialized means, variances and priors. */
            fs_use_initialized_cluster_means_variances_and_priors = TRUE;
            
            pso("Running EM for %d clusters.\n", num_clusters);
        }
        
    } while (num_clusters_annihilate > 0);
    
    ERE(get_target_vector(&log_sqrt_det_vp, num_clusters));
    for (cluster = 0; cluster < num_clusters; cluster++)
    {
        temp = 0.0;

        for (feature = 0; feature < num_features; feature++)
        {
            double var = var_mp->elements[ cluster ][ feature ]; 
            temp += SAFE_LOG(var);
        }

        log_sqrt_det_vp->elements[ cluster ] = temp / 2.0;
    }
    verify_vector(log_sqrt_det_vp);

    for (i = 0; i < fs_max_num_CEM_iterations; i++)
    {
        num_possible_split_merge_operations = num_clusters + ((num_clusters * (num_clusters - 1)) / 2);

        ERE(get_target_int_matrix(&split_merge_cluster_indices_mp, 2, num_possible_split_merge_operations));
        ERE(get_target_vector(&split_merge_probabilities_vp, num_possible_split_merge_operations));

        ERE(get_target_matrix(&log_cluster_distributions_mp, num_points, num_clusters));

        for (point = 0; point < num_points; point++)
        {
            ERE(get_matrix_row(&x_vp, feature_mp, point));

            for (cluster = 0; cluster < num_clusters; cluster++)
            {
                double log_a = SAFE_LOG(a_vp->elements[ cluster ]);

                double* x_ptr = x_vp->elements;
                double* u_ptr = u_mp->elements[ cluster ];
                double* v_ptr = var_mp->elements[ cluster ];

                double dev;
                double d = 0.0;

                for (feature = 0; feature < num_features; feature++)
                {
                    dev = *u_ptr - *x_ptr;

                    ASSERT_IS_NUMBER_DBL(dev); 
                    ASSERT_IS_FINITE_DBL(dev); 

                    temp = (dev / (*v_ptr)) * dev;

                    d += temp; 

                    ASSERT_IS_NUMBER_DBL(d); 
                    ASSERT_IS_FINITE_DBL(d); 

                    u_ptr++;
                    v_ptr++;
                    x_ptr++;
                }

                ASSERT_IS_NUMBER_DBL(d); 
                ASSERT_IS_FINITE_DBL(d); 

                log_cluster_distributions_mp->elements[ point ][ cluster ] = 0.0 - (0.5*d) - log_sqrt_det_vp->elements[ cluster ];
                (*local_P_mpp)->elements[ point ][ cluster ]  = log_a + log_cluster_distributions_mp->elements[ point ][ cluster ];

                ASSERT_IS_NUMBER_DBL(log_cluster_distributions_mp->elements[point][cluster]);
                ASSERT_IS_FINITE_DBL(log_cluster_distributions_mp->elements[point][cluster]);
                ASSERT_IS_NUMBER_DBL((*local_P_mpp)->elements[point][cluster]);
                ASSERT_IS_FINITE_DBL((*local_P_mpp)->elements[point][cluster]);
                
                log_cluster_distributions_mp->elements[ point ][ cluster ] -= ((double)num_features / 2.0) * SAFE_LOG(2.0 * M_PI); 

                ASSERT_IS_NUMBER_DBL(log_cluster_distributions_mp->elements[point][cluster]);
                ASSERT_IS_FINITE_DBL(log_cluster_distributions_mp->elements[point][cluster]);
            }
            
            ERE(get_matrix_row(&temp_vp, *local_P_mpp, point));
            temp = ow_exp_scale_by_sum_log_vector(temp_vp);
            ASSERT_IS_NUMBER_DBL(temp);
            ASSERT_IS_FINITE_DBL(temp);
            ERE(put_matrix_row(*local_P_mpp, temp_vp, point));
        }

        for (j = 0, n = num_clusters; j < num_clusters; j++)
        {
            /*
            ERE(get_matrix_col(&weighted_empirical_distribution_vp, *local_P_mpp, j));
            ERE(ow_scale_vector_by_sum(weighted_empirical_distribution_vp));

            ERE(get_matrix_col(&cluster_distribution_vp, log_cluster_distributions_mp, j));
            ow_exp_scale_by_sum_log_vector(cluster_distribution_vp);

            ERE(get_discrete_KL_divergence(weighted_empirical_distribution_vp, cluster_distribution_vp, &K_L_divergence));
            */

            ERE(get_matrix_col(&weighted_empirical_distribution_vp, *local_P_mpp, j));
            ERE(log_vector_2(&weighted_log_empirical_distribution_vp, weighted_empirical_distribution_vp, DBL_HALF_MOST_NEGATIVE));

            ERE(get_matrix_col(&log_cluster_distribution_vp, log_cluster_distributions_mp, j));
            
            ERE(get_discrete_KL_divergence_log_densities(weighted_log_empirical_distribution_vp, log_cluster_distribution_vp, &K_L_divergence));
            /* The following is a hack to deal with the fact that the
             * weighted empirical distribution is discrete and hence has an
             * out of bound differential entropy. This would ideally lead to
             * infinite KL divergence between it and a continuous
             * distribution. We convert it into a continuous distribution by 
             * spreading the point probability masses into very small
             * hypercubes around the discrete points.
            K_L_divergence = K_L_divergence + (DBL_HALF_MOST_POSITIVE / num_possible_split_merge_operations);
            */

            pso("KL divergence for splitting component %d with posterior %1.5f : %f.\n", j, a_vp->elements[j], K_L_divergence); 

            split_merge_cluster_indices_mp->elements[0][j] = j;
            split_merge_cluster_indices_mp->elements[1][j] = j;

            split_merge_probabilities_vp->elements[j] = K_L_divergence;

            for (k = j+1; k < num_clusters; k++)
            {
                split_merge_a = a_vp->elements[j] + a_vp->elements[k];

                ASSERT(split_merge_a > 0.0);

                ERE(get_zero_vector(&split_merge_u_vp, num_features));
                ERE(get_matrix_row(&u_1_vp, u_mp, j));
                ERE(ow_add_scalar_times_vector_to_vector(split_merge_u_vp, u_1_vp, (a_vp->elements[j]/split_merge_a)));
                ERE(get_matrix_row(&u_2_vp, u_mp, k));
                ERE(ow_add_scalar_times_vector_to_vector(split_merge_u_vp, u_2_vp, (a_vp->elements[k]/split_merge_a)));

                ERE(get_zero_vector(&split_merge_var_vp, num_features));
                ERE(get_matrix_row(&var_1_vp, var_mp, j));
                ERE(ow_add_scalar_times_vector_to_vector(split_merge_var_vp, var_1_vp, (a_vp->elements[j]/split_merge_a)));
                ERE(subtract_vectors(&var_1_vp, u_1_vp, split_merge_u_vp));
                ERE(ow_multiply_vectors(var_1_vp, var_1_vp));
                ERE(ow_add_scalar_times_vector_to_vector(split_merge_var_vp, var_1_vp, (a_vp->elements[j]/split_merge_a)));
                ERE(get_matrix_row(&var_2_vp, var_mp, k));
                ERE(ow_add_scalar_times_vector_to_vector(split_merge_var_vp, var_2_vp, (a_vp->elements[k]/split_merge_a)));
                ERE(subtract_vectors(&var_2_vp, u_2_vp, split_merge_u_vp));
                ERE(ow_multiply_vectors(var_2_vp, var_2_vp));
                ERE(ow_add_scalar_times_vector_to_vector(split_merge_var_vp, var_2_vp, (a_vp->elements[k]/split_merge_a)));
         
                /*
                   Though the following is theoretically correct, it is not
                   feasible to do in very high dimensions. This is because even
                   small values of variance in each dimension multiply to very
                   large values not allowed by the available precision. Instead,
                   do log(var) and then add.
                   
                db_rv(split_merge_var_vp);
                split_merge_log_sqrt_det = multiply_vector_elements(split_merge_var_vp);
                ASSERT_IS_NUMBER_DBL(split_merge_log_sqrt_det);
                ASSERT_IS_FINITE_DBL(split_merge_log_sqrt_det); 
                split_merge_log_sqrt_det = SAFE_LOG(split_merge_log_sqrt_det);
                */

                split_merge_log_sqrt_det = 0.0;

                for (feature = 0; feature < num_features; feature++)
                {
                    split_merge_log_sqrt_det += SAFE_LOG(split_merge_var_vp->elements[feature]);
                }

                ASSERT_IS_NUMBER_DBL(split_merge_log_sqrt_det);
                ASSERT_IS_FINITE_DBL(split_merge_log_sqrt_det); 
                split_merge_log_sqrt_det = split_merge_log_sqrt_det / 2.0;

                ERE(get_target_vector(&merge_log_cluster_distribution_vp, num_points));
                ERE(get_target_matrix(&merge_P_mp, num_points, (num_clusters-1)));

                for (point = 0; point < num_points; point++)
                {
                    ERE(get_matrix_row(&x_vp, feature_mp, point));

                    {
                        double log_a = SAFE_LOG(split_merge_a);
                        double* x_ptr = x_vp->elements;
                        double* u_ptr = split_merge_u_vp->elements;
                        double* v_ptr = split_merge_var_vp->elements;
                        double dev;
                        double d = 0.0;

                        for (feature = 0; feature < num_features; feature++)
                        {
                            dev = *u_ptr - *x_ptr;

                            ASSERT_IS_NUMBER_DBL(dev); 
                            ASSERT_IS_FINITE_DBL(dev); 

                            temp = (dev / (*v_ptr)) * dev;

                            d += temp; 

                            ASSERT_IS_NUMBER_DBL(d); 
                            ASSERT_IS_FINITE_DBL(d); 

                            u_ptr++;
                            v_ptr++;
                            x_ptr++;
                        }

                        ASSERT_IS_NUMBER_DBL(d); 
                        ASSERT_IS_FINITE_DBL(d); 

                        merge_log_cluster_distribution_vp->elements[point] = 0.0 - (0.5*d) - split_merge_log_sqrt_det;
                        merge_P_mp->elements[point][0] = log_a + merge_log_cluster_distribution_vp->elements[point];

                        ASSERT_IS_NUMBER_DBL(merge_log_cluster_distribution_vp->elements[point]);
                        ASSERT_IS_FINITE_DBL(merge_log_cluster_distribution_vp->elements[point]);
                        ASSERT_IS_NUMBER_DBL(merge_P_mp->elements[point][0]);
                        ASSERT_IS_FINITE_DBL(merge_P_mp->elements[point][0]);

                        merge_log_cluster_distribution_vp->elements[point] -= ((double)num_features / 2.0) * SAFE_LOG(2.0 * M_PI);

                        ASSERT_IS_NUMBER_DBL(merge_log_cluster_distribution_vp->elements[point]);
                        ASSERT_IS_FINITE_DBL(merge_log_cluster_distribution_vp->elements[point]);

                        for (l = 0, m = 1; l < num_clusters; l++)
                        {
                            if ( (l != j) && (l != k) )
                            {
                                log_a = SAFE_LOG(a_vp->elements[l]);

                                x_ptr = x_vp->elements;
                                u_ptr = u_mp->elements[l];
                                v_ptr = var_mp->elements[l];

                                d = 0.0;

                                for (feature = 0; feature < num_features; feature++)
                                {
                                    dev = *u_ptr - *x_ptr;

                                    ASSERT_IS_NUMBER_DBL(dev); 
                                    ASSERT_IS_FINITE_DBL(dev); 

                                    temp = (dev / (*v_ptr)) * dev;

                                    d += temp; 

                                    ASSERT_IS_NUMBER_DBL(d); 
                                    ASSERT_IS_FINITE_DBL(d); 

                                    u_ptr++;
                                    v_ptr++;
                                    x_ptr++;
                                }

                                ASSERT_IS_NUMBER_DBL(d); 
                                ASSERT_IS_FINITE_DBL(d); 

                                merge_P_mp->elements[point][m] = log_a - (0.5*d) - log_sqrt_det_vp->elements[l];

                                ASSERT_IS_NUMBER_DBL(merge_P_mp->elements[point][m]);
                                ASSERT_IS_FINITE_DBL(merge_P_mp->elements[point][m]);

                                m = m+1;
                            }
                        }

                        ERE(get_matrix_row(&temp_vp, merge_P_mp, point));
                        temp = ow_exp_scale_by_sum_log_vector(temp_vp);
                        ASSERT_IS_NUMBER_DBL(temp);
                        ASSERT_IS_FINITE_DBL(temp);
                        ERE(put_matrix_row(merge_P_mp, temp_vp, point));
                    }
                }

                /*
                ERE(get_matrix_col(&weighted_empirical_distribution_vp, merge_P_mp, 0));
                ERE(ow_scale_vector_by_sum(weighted_empirical_distribution_vp));

                ERE(copy_vector(&cluster_distribution_vp, merge_log_cluster_distribution_vp));
                ow_exp_scale_by_sum_log_vector(cluster_distribution_vp);

                ERE(get_discrete_KL_divergence(weighted_empirical_distribution_vp, cluster_distribution_vp, &K_L_divergence));
                */

                ERE(get_matrix_col(&weighted_empirical_distribution_vp, merge_P_mp, 0));
                ERE(log_vector_2(&weighted_log_empirical_distribution_vp, weighted_empirical_distribution_vp, DBL_HALF_MOST_NEGATIVE));

                ERE(copy_vector(&log_cluster_distribution_vp, merge_log_cluster_distribution_vp));

                ERE(get_discrete_KL_divergence_log_densities(weighted_log_empirical_distribution_vp, log_cluster_distribution_vp, &K_L_divergence)); 

                /* The following is a hack to deal with the fact that the
                 * weighted empirical distribution is discrete and hence has an
                 * out of bound differential entropy. This would ideally lead to
                 * infinite KL divergence between it and a continuous
                 * distribution. We convert it into a continuous distribution by 
                 * spreading the point probability masses into very small
                 * hypercubes around the discrete points.
                K_L_divergence = K_L_divergence + (DBL_HALF_MOST_POSITIVE / num_possible_split_merge_operations);
                */

                pso("KL divergence for merging components %d and %d : %f.\n", j, k, K_L_divergence);

                split_merge_cluster_indices_mp->elements[0][n] = j;
                split_merge_cluster_indices_mp->elements[1][n] = k;

                if (fs_force_equal_prob_for_CEM_split_and_merge == TRUE)
                {
                    split_merge_probabilities_vp->elements[n] = 1.0 / K_L_divergence;
                }
                else
                {
                    split_merge_probabilities_vp->elements[n] = beta / K_L_divergence;
                }

                n = n+1;
            }
        }

        ASSERT(n == num_possible_split_merge_operations);

        if (fs_force_equal_prob_for_CEM_split_and_merge == TRUE)
        {
            total_split_KL_divergence         = 0;
            total_merge_inverse_KL_divergence = 0;

            for (k = 0; k < num_clusters; k++)
            {
                total_split_KL_divergence = total_split_KL_divergence + split_merge_probabilities_vp->elements[k];
            }

            for (k = num_clusters; k < num_possible_split_merge_operations; k++)
            {
                total_merge_inverse_KL_divergence = total_merge_inverse_KL_divergence + split_merge_probabilities_vp->elements[k];
            }

            /* Compute beta online so that the total probability of all split
             * choices equals the total probability of all merge choices after
             * normalization. */ 
            beta = total_split_KL_divergence / total_merge_inverse_KL_divergence;

            for (k = num_clusters; k < num_possible_split_merge_operations; k++)
            {
                split_merge_probabilities_vp->elements[k] = beta * split_merge_probabilities_vp->elements[k];
            }
        }

        ERE(ow_scale_vector_by_sum(split_merge_probabilities_vp));
        db_rv(split_merge_probabilities_vp);

        ERE(get_zero_vector(&split_merge_cumulative_distribution_vp, num_possible_split_merge_operations));
        for (k = 0, temp = 0; k < num_possible_split_merge_operations; k++)
        {
            temp = temp + split_merge_probabilities_vp->elements[k];
            split_merge_cumulative_distribution_vp->elements[k] = temp;

            dbf(split_merge_cumulative_distribution_vp->elements[k]); 
        }

        ASSERT(split_merge_cumulative_distribution_vp->elements[num_possible_split_merge_operations-1] > 0.999);

        do
        {
            
            /* Operation sampling. */
            random_number = kjb_rand();
            for (k = 0; k < num_possible_split_merge_operations; k++)
            {
                if ( random_number <= split_merge_cumulative_distribution_vp->elements[k] )
                {
                    pso("Choosing %s operation with probability %f.\n",\
                            ((k < num_clusters) ? "split" : "merge"), split_merge_probabilities_vp->elements[k]);

                    break;
                }
            }

            ASSERT(k < num_possible_split_merge_operations);

            if (split_merge_cluster_indices_mp->elements[0][k] == split_merge_cluster_indices_mp->elements[1][k])
            {
                /* Split operation:
                 * Obtain the split means, variances and priors. */

                pso("Computing new parameters due to splitting component %d.\n",\
                        split_merge_cluster_indices_mp->elements[0][k]); 

                ERE(get_target_vector(&local_initial_a_vp, num_clusters+1));
                ERE(get_target_matrix(&local_initial_u_mp, num_clusters+1, num_features));
                ERE(get_target_matrix(&local_initial_var_mp, num_clusters+1, num_features));

                split_cluster_index = split_merge_cluster_indices_mp->elements[0][k];

                for (l = 0, m = 0; l < num_clusters; l++)
                {
                    if (l != split_cluster_index)
                    {
                        local_initial_a_vp->elements[m] = a_vp->elements[l];
                        ERE(copy_matrix_row(local_initial_u_mp, m, u_mp, l));
                        ERE(copy_matrix_row(local_initial_var_mp, m, var_mp, l));

                        m = m+1;
                    }
                }

                ERE(get_matrix_row(&temp_vp, var_mp, split_cluster_index));
                ERE(max_var_index = get_max_vector_element(temp_vp, &max_var));

                ERE(get_matrix_row(&u_1_vp, u_mp, split_cluster_index));
                ERE(get_matrix_row(&u_2_vp, u_mp, split_cluster_index));
                ERE(get_matrix_row(&var_1_vp, var_mp, split_cluster_index));
                ERE(get_matrix_row(&var_2_vp, var_mp, split_cluster_index));

                u_1_vp->elements[max_var_index] = u_1_vp->elements[max_var_index] - sqrt(max_var);
                u_2_vp->elements[max_var_index] = u_2_vp->elements[max_var_index] + sqrt(max_var);

                var_1_vp->elements[max_var_index] = 0.25 * max_var;
                var_2_vp->elements[max_var_index] = 0.25 * max_var;

                local_initial_a_vp->elements[m] = 0.5 * a_vp->elements[split_cluster_index];
                ERE(put_matrix_row(local_initial_u_mp, u_1_vp, m));
                ERE(put_matrix_row(local_initial_var_mp, var_1_vp, m));
                m = m+1;

                local_initial_a_vp->elements[m] = 0.5 * a_vp->elements[split_cluster_index];
                ERE(put_matrix_row(local_initial_u_mp, u_2_vp, m));
                ERE(put_matrix_row(local_initial_var_mp, var_2_vp, m));

                /* Hack: Store a copy of the standard user specified options 
                 * and call the EM module using initial means, variances and
                 * priors for one iteration. Then restore back the options.
                 */
                max_num_E_M_iterations = fs_max_num_iterations;
                use_initialization_E_M = fs_use_initialized_cluster_means_variances_and_priors; 

                /*
                   fs_max_num_iterations = 1;
                 */
                fs_use_initialized_cluster_means_variances_and_priors = TRUE;

                pso("Determining log likelihood due to split resulting in %d clusters.\n", num_clusters+1);
                ERE(new_num_clusters = get_independent_GMM_3((num_clusters+1), 
                            feature_mp,
                            (Vector*) NULL,
                            local_initial_a_vp, 
                            local_initial_u_mp,
                            local_initial_var_mp,
                            (Vector**) NULL, 
                            (Matrix**) NULL,
                            (Matrix**) NULL,  
                            (Matrix**) NULL, 
                            &new_log_likelihood,
                            &new_held_out_log_likelihood,
                            &num_E_M_iterations));                

                fs_max_num_iterations = max_num_E_M_iterations;
                fs_use_initialized_cluster_means_variances_and_priors = use_initialization_E_M;
            }
            else
            {
                /* Merge operation:
                 * Obtain the merged means, variances and priors. */

                pso("Computing new parameters due to merging components %d and %d.\n",\
                        split_merge_cluster_indices_mp->elements[0][k], split_merge_cluster_indices_mp->elements[1][k]); 

                ERE(get_target_vector(&local_initial_a_vp, num_clusters-1));
                ERE(get_target_matrix(&local_initial_u_mp, num_clusters-1, num_features));
                ERE(get_target_matrix(&local_initial_var_mp, num_clusters-1, num_features));

                merge_cluster_index_1 = split_merge_cluster_indices_mp->elements[0][k];
                merge_cluster_index_2 = split_merge_cluster_indices_mp->elements[1][k];

                for (l = 0, m = 0; l < num_clusters; l++)
                {
                    if ( (l != merge_cluster_index_1) && (l != merge_cluster_index_2) )
                    {
                        local_initial_a_vp->elements[m] = a_vp->elements[l];
                        ERE(copy_matrix_row(local_initial_u_mp, m, u_mp, l));
                        ERE(copy_matrix_row(local_initial_var_mp, m, var_mp, l));

                        m = m+1;
                    }
                }

                split_merge_a = a_vp->elements[merge_cluster_index_1] + a_vp->elements[merge_cluster_index_2];

                ASSERT(split_merge_a > 0.0);

                ERE(get_zero_vector(&split_merge_u_vp, num_features));
                ERE(get_matrix_row(&u_1_vp, u_mp, merge_cluster_index_1));
                ERE(ow_add_scalar_times_vector_to_vector(split_merge_u_vp, u_1_vp, (a_vp->elements[merge_cluster_index_1]/split_merge_a)));
                ERE(get_matrix_row(&u_2_vp, u_mp, merge_cluster_index_2));
                ERE(ow_add_scalar_times_vector_to_vector(split_merge_u_vp, u_2_vp, (a_vp->elements[merge_cluster_index_2]/split_merge_a)));

                ERE(get_zero_vector(&split_merge_var_vp, num_features));
                ERE(get_matrix_row(&var_1_vp, var_mp, merge_cluster_index_1));
                ERE(ow_add_scalar_times_vector_to_vector(split_merge_var_vp, var_1_vp, (a_vp->elements[merge_cluster_index_1]/split_merge_a)));
                ERE(subtract_vectors(&var_1_vp, u_1_vp, split_merge_u_vp));
                ERE(ow_multiply_vectors(var_1_vp, var_1_vp));
                ERE(ow_add_scalar_times_vector_to_vector(split_merge_var_vp, var_1_vp, (a_vp->elements[merge_cluster_index_1]/split_merge_a)));
                ERE(get_matrix_row(&var_2_vp, var_mp, merge_cluster_index_2));
                ERE(ow_add_scalar_times_vector_to_vector(split_merge_var_vp, var_2_vp, (a_vp->elements[merge_cluster_index_2]/split_merge_a)));
                ERE(subtract_vectors(&var_2_vp, u_2_vp, split_merge_u_vp));
                ERE(ow_multiply_vectors(var_2_vp, var_2_vp));
                ERE(ow_add_scalar_times_vector_to_vector(split_merge_var_vp, var_2_vp, (a_vp->elements[merge_cluster_index_2]/split_merge_a)));

                local_initial_a_vp->elements[m] = split_merge_a;
                ERE(put_matrix_row(local_initial_u_mp, split_merge_u_vp, m));
                ERE(put_matrix_row(local_initial_var_mp, split_merge_var_vp, m)); 

                /* Hack: Store a copy of the standard user specified options 
                 * and call the EM module using initial means, variances and
                 * priors for one iteration. Then restore back the options.
                 */
                max_num_E_M_iterations = fs_max_num_iterations;
                use_initialization_E_M = fs_use_initialized_cluster_means_variances_and_priors; 

                /*
                   fs_max_num_iterations = 1;
                 */
                fs_use_initialized_cluster_means_variances_and_priors = TRUE;

                pso("Determining log likelihood due to merge resulting in %d clusters.\n", num_clusters-1);
                ERE(new_num_clusters = get_independent_GMM_3((num_clusters-1), 
                            feature_mp,
                            (Vector*) NULL,
                            local_initial_a_vp, 
                            local_initial_u_mp,
                            local_initial_var_mp,
                            (Vector**) NULL, 
                            (Matrix**) NULL, 
                            (Matrix**) NULL,  
                            (Matrix**) NULL, 
                            &new_log_likelihood,
                            &new_held_out_log_likelihood,
                            &num_E_M_iterations));                

                fs_max_num_iterations = max_num_E_M_iterations;
                fs_use_initialized_cluster_means_variances_and_priors = use_initialization_E_M;
            }

            /* Acceptance sampling. */
            adjusted_log_likelihood = log_likelihood - (0.5 * num_clusters * SAFE_LOG(num_points / 12.0)); 
            adjusted_log_likelihood = adjusted_log_likelihood - (0.5 * num_clusters * (num_parameters_per_cluster + 1.0));
            for (cluster = 0; cluster < num_clusters; cluster++)
            {
                adjusted_log_likelihood = adjusted_log_likelihood - (0.5 * num_parameters_per_cluster * SAFE_LOG(num_points * a_vp->elements[cluster] / 12.0));
            }

            /* Keep track of the best model based on maximum adjusted log
             * likelihood. */
            if (adjusted_log_likelihood > max_adjusted_log_likelihood)
            {
                max_adjusted_log_likelihood = adjusted_log_likelihood;

                if (a_vpp != NULL)
                {
                    ERE(copy_vector(a_vpp, a_vp));
                }

                if (u_mpp != NULL)
                {
                    ERE(copy_matrix(u_mpp, u_mp));
                }

                if (var_mpp != NULL)
                {
                    ERE(copy_matrix(var_mpp, var_mp));
                }

                if (P_mpp != NULL)
                {
                    ERE(copy_matrix(P_mpp, local_P_mp));
                }

                /* Keep track of intermediate results. */
                if (fs_write_CEM_intermediate_results == TRUE)
                {
                    if ( (a_vp != NULL) && (u_mp != NULL) && (var_mp != NULL) )
                    {
                        ERE(get_target_matrix(&intermediate_a_mvp->elements[intermediate_result_index], num_clusters, 1));
                        ERE(get_target_matrix(&intermediate_u_mvp->elements[intermediate_result_index], num_clusters, num_features));
                        ERE(get_target_matrix(&intermediate_var_mvp->elements[intermediate_result_index], num_clusters, num_features));

                        intermediate_a_mp = intermediate_a_mvp->elements[intermediate_result_index];
                        intermediate_u_mp = intermediate_u_mvp->elements[intermediate_result_index];
                        intermediate_var_mp = intermediate_var_mvp->elements[intermediate_result_index];

                        for (cluster = 0; cluster < num_clusters; cluster++)
                        {
                            intermediate_a_mp->elements[cluster][0] = a_vp->elements[cluster];
                            ERE(copy_matrix_row(intermediate_u_mp, cluster, u_mp, cluster));
                            ERE(copy_matrix_row(intermediate_var_mp, cluster, var_mp, cluster));
                        }

                        intermediate_result_index = intermediate_result_index + 1; 
                    }
                }
            }

            adjusted_new_log_likelihood = new_log_likelihood - (0.5 * new_num_clusters * SAFE_LOG(num_points / 12.0)); 
            adjusted_new_log_likelihood = adjusted_new_log_likelihood - (0.5 * new_num_clusters * (num_parameters_per_cluster + 1.0));
            for (cluster = 0; cluster < new_num_clusters; cluster++)
            {
                adjusted_new_log_likelihood = adjusted_new_log_likelihood - (0.5 * num_parameters_per_cluster * SAFE_LOG(num_points * local_initial_a_vp->elements[cluster] / 12.0));
            }

            pso("%-3d: Old log likelihood with %d clusters: %12e.\n", i+1, num_clusters, log_likelihood);
            pso("%-3d: New log likelihood with %d clusters: %12e.\n", i+1, new_num_clusters, new_log_likelihood);
            pso("%-3d: Adjusted old log likelihood with %d clusters: %12e.\n", i+1, num_clusters, adjusted_log_likelihood);
            pso("%-3d: Adjusted new log likelihood with %d clusters: %12e.\n", i+1, new_num_clusters, adjusted_new_log_likelihood);

            acceptance_prob = exp((adjusted_new_log_likelihood - adjusted_log_likelihood)/gamma);

            if (kjb_rand() < acceptance_prob)
            {
                break;
            }

            i = i+1;

            if (i >= fs_max_num_CEM_iterations)
            {
                break;
            }
            
        } while (1);
        
        if (i >= fs_max_num_CEM_iterations)
        {
            break;
        }
        
        /*
        if (kjb_rand() < acceptance_prob)
        */
        {
            pso("Accepting %s operation with probability %10e.\n",\
            ((split_merge_cluster_indices_mp->elements[0][k] ==\
            split_merge_cluster_indices_mp->elements[1][k]) ? "split" :\
            "merge"), acceptance_prob);

            /* Keep track of intermediate results. */
            if (fs_write_CEM_intermediate_results == TRUE)
            {
                if ( (local_initial_a_vp != NULL) && (local_initial_u_mp != NULL) && (local_initial_var_mp != NULL) )
                {
                    ERE(get_target_matrix(&intermediate_a_mvp->elements[intermediate_result_index], new_num_clusters, 1));
                    ERE(get_target_matrix(&intermediate_u_mvp->elements[intermediate_result_index], new_num_clusters, num_features));
                    ERE(get_target_matrix(&intermediate_var_mvp->elements[intermediate_result_index], new_num_clusters, num_features));

                    intermediate_a_mp = intermediate_a_mvp->elements[intermediate_result_index];
                    intermediate_u_mp = intermediate_u_mvp->elements[intermediate_result_index];
                    intermediate_var_mp = intermediate_var_mvp->elements[intermediate_result_index];

                    for (cluster = 0; cluster < new_num_clusters; cluster++)
                    {
                        intermediate_a_mp->elements[cluster][0] = local_initial_a_vp->elements[cluster];
                        ERE(copy_matrix_row(intermediate_u_mp, cluster, local_initial_u_mp, cluster));
                        ERE(copy_matrix_row(intermediate_var_mp, cluster, local_initial_var_mp, cluster));
                    }

                    intermediate_result_index = intermediate_result_index + 1; 
                }
            }

            do
            {
                /* Hack: Store a copy of the standard user specified EM
                 * initialization option and call the EM module using initial 
                 * means, variances and priors. Then restore back the options.
                 */
                use_initialization_E_M = fs_use_initialized_cluster_means_variances_and_priors; 

                fs_use_initialized_cluster_means_variances_and_priors = TRUE;

                ERE(num_clusters = get_independent_GMM_3(new_num_clusters, 
                                                         feature_mp,
                                                         (Vector*) NULL,
                                                         local_initial_a_vp, 
                                                         local_initial_u_mp,
                                                         local_initial_var_mp,
                                                         &a_vp, &u_mp, &var_mp,  
                                                         local_P_mpp, 
                                                         &log_likelihood,
                                                         &held_out_log_likelihood,
                                                         &num_E_M_iterations));                

                fs_use_initialized_cluster_means_variances_and_priors = use_initialization_E_M;

                pso("%-3d: Log likelihood with %d clusters after EM: %12e.\n", i+1, num_clusters, log_likelihood);

                /* Keep track of intermediate results. */
                if (fs_write_CEM_intermediate_results == TRUE)
                {
                    if ( (a_vp != NULL) && (u_mp != NULL) && (var_mp != NULL) )
                    {
                        ERE(get_target_matrix(&intermediate_a_mvp->elements[intermediate_result_index], num_clusters, 1));
                        ERE(get_target_matrix(&intermediate_u_mvp->elements[intermediate_result_index], num_clusters, num_features));
                        ERE(get_target_matrix(&intermediate_var_mvp->elements[intermediate_result_index], num_clusters, num_features));

                        intermediate_a_mp = intermediate_a_mvp->elements[intermediate_result_index];
                        intermediate_u_mp = intermediate_u_mvp->elements[intermediate_result_index];
                        intermediate_var_mp = intermediate_var_mvp->elements[intermediate_result_index];

                        for (cluster = 0; cluster < num_clusters; cluster++)
                        {
                            intermediate_a_mp->elements[cluster][0] = a_vp->elements[cluster];
                            ERE(copy_matrix_row(intermediate_u_mp, cluster, u_mp, cluster));
                            ERE(copy_matrix_row(intermediate_var_mp, cluster, var_mp, cluster));
                        }

                        intermediate_result_index = intermediate_result_index + 1; 
                    }
                }

                /* Cluster annihilation. */
                num_clusters_annihilate = 0;
                total_mass_annihilate   = 0.0;
                for (cluster = 0; cluster < num_clusters; cluster++)
                {
                    if ( (a_vp->elements[cluster] * num_points) < num_parameters_per_cluster )
                    {
                        ERE(remove_matrix_row((Vector**) NULL, u_mp, (cluster - num_clusters_annihilate)));
                        ERE(remove_matrix_row((Vector**) NULL, var_mp, (cluster - num_clusters_annihilate)));
                        
                        pso("Annihilating cluster %d with posterior %f.\n", cluster, a_vp->elements[cluster]);
                        num_clusters_annihilate = num_clusters_annihilate + 1;
                        total_mass_annihilate = total_mass_annihilate + a_vp->elements[cluster];
                    }
                }

                ERE(get_target_vector(&temp_vp, (num_clusters - num_clusters_annihilate)));
                for (cluster = 0, m = 0; cluster < num_clusters; cluster++)
                {
                    if ( (a_vp->elements[cluster] * num_points) >= num_parameters_per_cluster )
                    {
                        temp_vp->elements[m] = a_vp->elements[cluster] + (total_mass_annihilate/(num_clusters - num_clusters_annihilate));
                        m = m+1;
                    }
                }
                ERE(scale_vector_by_sum(&a_vp, temp_vp));

                num_clusters     = num_clusters - num_clusters_annihilate;
                new_num_clusters = num_clusters;

                if (num_clusters_annihilate > 0)
                {
                    ERE(copy_vector(&local_initial_a_vp, a_vp));
                    ERE(copy_matrix(&local_initial_u_mp, u_mp));
                    ERE(copy_matrix(&local_initial_var_mp, var_mp));

                    /* Keep track of intermediate results. */
                    if (fs_write_CEM_intermediate_results == TRUE)
                    {
                        if ( (a_vp != NULL) && (u_mp != NULL) && (var_mp != NULL) )
                        {
                            ERE(get_target_matrix(&intermediate_a_mvp->elements[intermediate_result_index], num_clusters, 1));
                            ERE(get_target_matrix(&intermediate_u_mvp->elements[intermediate_result_index], num_clusters, num_features));
                            ERE(get_target_matrix(&intermediate_var_mvp->elements[intermediate_result_index], num_clusters, num_features));

                            intermediate_a_mp = intermediate_a_mvp->elements[intermediate_result_index];
                            intermediate_u_mp = intermediate_u_mvp->elements[intermediate_result_index];
                            intermediate_var_mp = intermediate_var_mvp->elements[intermediate_result_index];

                            for (cluster = 0; cluster < num_clusters; cluster++)
                            {
                                intermediate_a_mp->elements[cluster][0] = a_vp->elements[cluster];
                                ERE(copy_matrix_row(intermediate_u_mp, cluster, u_mp, cluster));
                                ERE(copy_matrix_row(intermediate_var_mp, cluster, var_mp, cluster));
                            }

                            intermediate_result_index = intermediate_result_index + 1; 
                        }
                    }

                    pso("Running EM for %d clusters.\n", num_clusters);
                }

            } while (num_clusters_annihilate > 0);
            
            ERE(get_target_vector(&log_sqrt_det_vp, num_clusters));
            for (cluster = 0; cluster < num_clusters; cluster++)
            {
                temp = 0.0;

                for (feature = 0; feature < num_features; feature++)
                {
                    double var = var_mp->elements[ cluster ][ feature ]; 
                    temp += SAFE_LOG(var);
                }

                log_sqrt_det_vp->elements[ cluster ] = temp / 2.0;
            }
            verify_vector(log_sqrt_det_vp);
        }
    }

    adjusted_log_likelihood = log_likelihood - (0.5 * num_clusters * SAFE_LOG(num_points / 12.0)); 
    adjusted_log_likelihood = adjusted_log_likelihood - (0.5 * num_clusters * (num_parameters_per_cluster + 1.0));
    for (cluster = 0; cluster < num_clusters; cluster++)
    {
        adjusted_log_likelihood = adjusted_log_likelihood - (0.5 * num_parameters_per_cluster * SAFE_LOG(num_points * a_vp->elements[cluster] / 12.0));
    }

    if (adjusted_log_likelihood > max_adjusted_log_likelihood)
    {
        max_adjusted_log_likelihood = adjusted_log_likelihood;

        if (a_vpp != NULL)
        {
            ERE(copy_vector(a_vpp, a_vp));
        }

        if (u_mpp != NULL)
        {
            ERE(copy_matrix(u_mpp, u_mp));
        }

        if (var_mpp != NULL)
        {
            ERE(copy_matrix(var_mpp, var_mp));
        }

        if (P_mpp != NULL)
        {
            ERE(copy_matrix(P_mpp, local_P_mp));
        }

        /* Keep track of intermediate results. */
        if (fs_write_CEM_intermediate_results == TRUE)
        {
            if ( (a_vp != NULL) && (u_mp != NULL) && (var_mp != NULL) )
            {
                ERE(get_target_matrix(&intermediate_a_mvp->elements[intermediate_result_index], num_clusters, 1));
                ERE(get_target_matrix(&intermediate_u_mvp->elements[intermediate_result_index], num_clusters, num_features));
                ERE(get_target_matrix(&intermediate_var_mvp->elements[intermediate_result_index], num_clusters, num_features));

                intermediate_a_mp = intermediate_a_mvp->elements[intermediate_result_index];
                intermediate_u_mp = intermediate_u_mvp->elements[intermediate_result_index];
                intermediate_var_mp = intermediate_var_mvp->elements[intermediate_result_index];

                for (cluster = 0; cluster < num_clusters; cluster++)
                {
                    intermediate_a_mp->elements[cluster][0] = a_vp->elements[cluster];
                    ERE(copy_matrix_row(intermediate_u_mp, cluster, u_mp, cluster));
                    ERE(copy_matrix_row(intermediate_var_mp, cluster, var_mp, cluster));
                }

                intermediate_result_index = intermediate_result_index + 1; 
            }
        }
    }
        
    if (fs_normalize_data)
    {
        ERE(ow_un_normalize_cluster_data((u_mpp == NULL) ? NULL : *u_mpp, 
                                         (var_mpp == NULL) ? NULL : *var_mpp, 
                                         feature_mean_vp, 
                                         feature_var_vp, 
                                         norm_stdev)); 
    }

    /* Copying the final result at the end. */
    if (fs_write_CEM_intermediate_results == TRUE)
    {
        if ( (a_vpp != NULL) && (u_mpp != NULL) && (var_mpp != NULL) )
        {
            if ( (*a_vpp != NULL) && (*u_mpp != NULL) && (*var_mpp != NULL) )
            {
                ERE(get_target_matrix(&intermediate_a_mvp->elements[intermediate_result_index], (*a_vpp)->length, 1));
                ERE(get_target_matrix(&intermediate_u_mvp->elements[intermediate_result_index], (*a_vpp)->length, num_features));
                ERE(get_target_matrix(&intermediate_var_mvp->elements[intermediate_result_index], (*a_vpp)->length, num_features));

                intermediate_a_mp = intermediate_a_mvp->elements[intermediate_result_index];
                intermediate_u_mp = intermediate_u_mvp->elements[intermediate_result_index];
                intermediate_var_mp = intermediate_var_mvp->elements[intermediate_result_index];

                for (cluster = 0; cluster < (*a_vpp)->length; cluster++)
                {
                    intermediate_a_mp->elements[cluster][0] = (*a_vpp)->elements[cluster];
                    ERE(copy_matrix_row(intermediate_u_mp, cluster, *u_mpp, cluster));
                    ERE(copy_matrix_row(intermediate_var_mp, cluster, *var_mpp, cluster));
                }

                intermediate_result_index = intermediate_result_index + 1; 
            }
        }
    }

    /* Write all the intermediate and final results to disk. */
    if (fs_write_CEM_intermediate_results == TRUE)
    {
        ERE(write_matrix_vector(intermediate_a_mvp, "intermediate_a_mvp"));
        ERE(write_matrix_vector(intermediate_u_mvp, "intermediate_u_mvp"));
        ERE(write_matrix_vector(intermediate_var_mvp, "intermediate_var_mvp"));
    }
    
    free_vector(a_vp);
    free_matrix(u_mp);
    free_matrix(var_mp);
    free_matrix(local_P_mp);

    free_matrix(norm_feature_mp);
    free_vector(feature_var_vp); 
    free_vector(feature_mean_vp); 

    free_matrix(perturbed_feature_mp);

    free_int_matrix(split_merge_cluster_indices_mp);
    free_vector(split_merge_probabilities_vp);
    free_vector(split_merge_cumulative_distribution_vp);
    free_vector(log_sqrt_det_vp);
    free_vector(x_vp);
    free_matrix(log_cluster_distributions_mp);
    free_vector(weighted_empirical_distribution_vp);
    free_vector(weighted_log_empirical_distribution_vp);
    free_vector(cluster_distribution_vp);
    free_vector(log_cluster_distribution_vp);

    free_vector(split_merge_u_vp);
    free_vector(split_merge_var_vp);

    free_vector(u_1_vp);
    free_vector(u_2_vp);
    free_vector(var_1_vp);
    free_vector(var_2_vp);

    free_vector(merge_log_cluster_distribution_vp);
    free_matrix(merge_P_mp);
    free_vector(temp_vp);

    free_vector(local_initial_a_vp);
    free_matrix(local_initial_u_mp);
    free_matrix(local_initial_var_mp);
    
    free_matrix_vector(intermediate_a_mvp);
    free_matrix_vector(intermediate_u_mvp);
    free_matrix_vector(intermediate_var_mvp);

    /*
    free_matrix(intermediate_a_mp);
    free_matrix(intermediate_u_mp);
    free_matrix(intermediate_var_mp);
    */

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * Kobus: This changes true_log_distribution_vp and assumed_log_distribution_vp
 * due to the call to normalization. 
*/
static int get_discrete_KL_divergence_log_densities
(
    Vector* true_log_distribution_vp,
    Vector* assumed_log_distribution_vp,
    double*       K_L_divergence_ptr
)
{
    int num_values;
    int i;

    Vector* true_distribution_vp = NULL;

    if (true_log_distribution_vp == NULL)
    {
        set_bug("True distribution vector is NULL.");
        return ERROR;
    }

    if (assumed_log_distribution_vp == NULL)
    {
        set_bug("Assumed distribution vector is NULL.");
        return ERROR;
    }
    
    num_values = true_log_distribution_vp->length;
    
    ASSERT(num_values == assumed_log_distribution_vp->length);

    *K_L_divergence_ptr = 0;
    
    /* Making sure the input log densities are normalized */

    /* It is only necessary to make sure the true distribution is normalized.
     * This is because the assumed distribution might be in a continuous space
     * in which case the sum of the point density values over the input points
     * need not sum to 1. */
    ERE(ow_normalize_log_prob_vp(true_log_distribution_vp));
    ERE(ow_normalize_log_prob_vp(assumed_log_distribution_vp)); 

    ERE(copy_vector(&true_distribution_vp, true_log_distribution_vp));
    ow_exp_scale_by_sum_log_vector(true_distribution_vp);

    for (i = 0; i < num_values; i++)
    {
        /* Sanity checks */
        ASSERT(true_distribution_vp->elements[i] >= 0.0);

        ASSERT_IS_NUMBER_DBL(true_log_distribution_vp->elements[i]);
        ASSERT_IS_NUMBER_DBL(assumed_log_distribution_vp->elements[i]);
    
        /*
        ASSERT_IS_FINITE_DBL(true_log_distribution_vp->elements[i]);
        */

        ASSERT_IS_FINITE_DBL(assumed_log_distribution_vp->elements[i]);

        if (assumed_log_distribution_vp->elements[i] <= DBL_HALF_MOST_NEGATIVE)
        {
            /*
            *K_L_divergence_ptr = 1000;
            return NO_ERROR;
            */

            /*
            set_bug("Support of the assumed distribution is smaller than that of the true distribution.");
            return ERROR;
            */
        }
        else if (true_distribution_vp->elements[i] == 0)
        {
            *K_L_divergence_ptr = *K_L_divergence_ptr + 0;
        }
        else
        {
            *K_L_divergence_ptr = *K_L_divergence_ptr + (true_distribution_vp->elements[i] * (true_log_distribution_vp->elements[i] - assumed_log_distribution_vp->elements[i]));
        }
    }
    
    free_vector(true_distribution_vp);

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *            get_GMM_blk_compound_sym_cov
 *                               
 * Finds a Gaussian mixture model (GMM) where the Gaussians have block compound
 * symmetrical covariances with shared parameters. Specifically, each feature
 * has the same mean (mu) and variance (sig^2 + tau^2). The covariance
 * between any pair of features is either (tau^2) or 0. This holds for all the
 * Gaussians in the mixture. However the block diagonal structure of the
 * covariance matrices of the different Gaussians in the mixture is different.
 * Further it is assumed that the ratio (tau^2 / sig^2) is known beforehand.
 * This amounts to reducing one degree of freedom in fitting the model but it
 * enables closed form solutions for optimum parameter values (mu, sig^2) in
 * the M-step of EM. A suitable value for the ratio can be estimated using
 * cross-validation.
 *
 * In particular, it fits:
 * |         p(x) = sum  a-sub-i *  g(mu-vec, cov-sub-i, x)
 * |                 i
 * where a-sub-i is the prior probability for the mixuture compoenent (cluster),
 * mu-vec is the mean vector with all components equal to mu, cov-sub-i is the
 * covariance matrix for the component i, and g(mu,cov,x) is a Gaussian with
 * mean mu and covariance cov. 
 *
 * The data matrix feature_mp is an N by M matrix where N is the number of data
 * points, and M is the number of features. 
 *
 * The argument block_diag_sizes_vvp specifies the block diagonal structures of
 * the covariances of the Gaussian components. The number of vectors in this
 * argument is equal to the number of mixture compoenents (clusters), K. Each
 * vector is a list of sizes of the block diagonals from top to bottom in the
 * corresponding covariance matrix. For eg., the vector corresponding to a
 * Gaussian component with all independent features would consist of 1 as all
 * its elements and the number of elements in the vector equal to M. And a
 * Gaussian component with two block diagonals of the same size in its
 * covariance matrix would be specified by a vector with two elements, each
 * equal to M/2.
 *
 * The argument tau_sig_sqr specifies the ratio as mentioned above.
 *
 * initial_a_vp, initial_mu and initial_sig_sqr can be used to specify the
 * initial values of the parameters for EM.
 *
 * The model parameters are put into *a_vpp, *mu_ptr, and *sig_sqr_ptr. Any
 * of a_vpp, mu_ptr, or sig_sqr_ptr is NULL if that value is not needed.
 *
 * If P_mpp is not NULL, then the soft clustering (cluster membership) for each
 * data point is returned. In that case, *P_mpp will be N by K. 
 *
 * Returns :
 *    If the routine fails (due to storage allocation), then ERROR is returned
 *    with an error message being set. Otherwise NO_ERROR is returned. 
 *
 * Index: clustering, EM, GMM
 *
 * Note: The covariance structure assumed in this routine is often referred to
 * as "block compound-symmetry" structure especially in the mixed-models ANOVA
 * literature. It is useful in modeling data with repeated measures in ANOVA
 * using mixed-models. For example see:
 * http://www.asu.edu/sas/sasdoc/sashtml/stat/chap41/sect23.htm
 * http://www.tufts.edu/~gdallal/repeat2.htm 
 *
 * Authors: Prasad Gabbur, Kobus Barnard.
 * -----------------------------------------------------------------------------
*/

int get_GMM_blk_compound_sym_cov
(
    const Int_vector_vector* block_diag_sizes_vvp,
    const Matrix*            feature_mp,
    const Vector*            held_out_indicator_vp,
    const Vector*            initial_a_vp,
    const double             initial_mu,
    const double             initial_sig_sqr,
    const double             tau_sig_sqr_ratio,
    Vector**                 a_vpp, 
    double*                  mu_ptr,
    double*                  sig_sqr_ptr,
    Matrix**                 P_mpp,
    double*                  log_likelihood_ptr,
    double*                  held_out_log_likelihood_ptr,
    int*                     num_iterations_ptr
)
{
    int num_clusters = block_diag_sizes_vvp->length;
    int num_points   = feature_mp->num_rows;
    int num_features = feature_mp->num_cols;
   
    Int_vector*  block_diag_sizes_vp = NULL;
    
    int      cluster;
    double   var_offset = fs_var_offset; 
    int      it, i, j, k, l;
    double   mu, sig_sqr;
    double   new_mu, new_sig_sqr;
    Vector*  I_vp          = NULL;
    Vector*  a_vp          = NULL;
    Vector*  p_sum_vp      = NULL;
    Vector*  x_vp          = NULL;
    Vector*  u_vp          = NULL;
    Vector*  centered_x_vp = NULL;
    Vector*  ones_vp       = NULL;
    Vector*  temp_vp       = NULL;
    double   log_likelihood               = DBL_HALF_MOST_NEGATIVE;
    double   prev_log_likelihood          = DBL_HALF_MOST_NEGATIVE;
    double   held_out_log_likelihood      = DBL_HALF_MOST_NEGATIVE;
    double   prev_held_out_log_likelihood = DBL_HALF_MOST_NEGATIVE;
    double   diff, held_out_diff;

    Matrix_vector* S_tilda_inv_mvp = NULL;
    Matrix*        S_tilda_inv_mp  = NULL;
    Matrix*        S_inv_mp        = NULL;

    double zero_order_stat;
    double first_order_stat;
    double second_order_stat;

    int     result = NO_ERROR;
#ifdef DONT_USE_GAUSS_ROUTINE
    Vector* log_sqrt_det_vp = NULL;
#endif 

#ifdef DONT_USE_GAUSS_ROUTINE
    double temp;
#endif
    double p, d;
    double log_a;
    int    num_block_diags;
    int    block_diag_size;
    int    sum_block_sizes;

    if (P_mpp != NULL)
    {
        ERE(get_target_matrix(P_mpp, num_points, num_clusters));
    }

    if (held_out_indicator_vp != NULL)
    {
        if (held_out_indicator_vp->length != num_points)
        {
            result = ERROR;
        }
    }
            
    if (    (get_target_vector(&I_vp, num_clusters) == ERROR)
         || (get_target_vector(&a_vp, num_clusters) == ERROR) 
#ifdef DONT_USE_GAUSS_ROUTINE
         || (get_target_vector(&log_sqrt_det_vp, num_clusters) == ERROR) 
#endif 
       )
    {
        result = ERROR;
    }

    dbi(num_clusters); 

    if (result != ERROR)
    {
        result = get_target_matrix_vector(&S_tilda_inv_mvp, num_clusters);
    }

    for (cluster = 0; cluster < num_clusters; cluster++)
    {
        if (result == ERROR) { NOTE_ERROR(); break; }

        sum_block_sizes = 0;

        /*
        S_tilda_inv_mp = S_tilda_inv_mvp->elements[ cluster ];

        result = get_zero_matrix(&S_tilda_inv_mp, num_features, num_features);
        if (result == ERROR) { NOTE_ERROR(); break; }
        */

        result = get_zero_matrix(&(S_tilda_inv_mvp->elements[ cluster ]), num_features, num_features);
        if (result == ERROR) { NOTE_ERROR(); break; }
        
        S_tilda_inv_mp = S_tilda_inv_mvp->elements[ cluster ];

        block_diag_sizes_vp = block_diag_sizes_vvp->elements[ cluster ];

        db_irv(block_diag_sizes_vp);

        num_block_diags = block_diag_sizes_vp->length;

        for (i = 0, j = 0, k = 0; i < num_block_diags; i++)
        {
            block_diag_size = block_diag_sizes_vp->elements[i];

            sum_block_sizes = sum_block_sizes + block_diag_size;

            l = j;

            for (j = l; j < (l + block_diag_size); j++)
            {
                for (k = l; k < (l + block_diag_size); k++)
                {
                    if (j == k)
                    {
                        S_tilda_inv_mp->elements[j][k] = 1 + ((block_diag_size - 1) * tau_sig_sqr_ratio);
                        S_tilda_inv_mp->elements[j][k] = S_tilda_inv_mp->elements[j][k] / \
                                                         (1 + (block_diag_size * tau_sig_sqr_ratio));
                    }
                    else
                    {
                        S_tilda_inv_mp->elements[j][k] = -tau_sig_sqr_ratio / \
                                                         (1 + (block_diag_size * tau_sig_sqr_ratio));
                    }
                }
            }
        }

        if (sum_block_sizes != num_features)
        {
            set_bug("Block diagonal sizes for cluster %d do not sum \
                    to the feature dimensionality:", (cluster + 1));
            return ERROR;
        }
    }

    for (it = 0; it < fs_max_num_iterations; it++)
    {
        log_likelihood = 0.0;

        held_out_log_likelihood = 0.0;

        if (result == ERROR) { NOTE_ERROR(); break; }

        result = get_zero_vector(&p_sum_vp, num_clusters);
        if (result == ERROR) { NOTE_ERROR(); break; }

        new_mu            = 0.0;
        new_sig_sqr  = 0.0;
        zero_order_stat   = 0.0;
        first_order_stat  = 0.0;
        second_order_stat = 0.0;

        if ( (it == 0) && (fs_use_initialized_cluster_means_variances_and_priors == TRUE) )
        {
            mu = initial_mu;

            sig_sqr = initial_sig_sqr + var_offset; 

            if (initial_a_vp != NULL)
            {
                ERE(get_target_vector(&a_vp, num_clusters));

                for (cluster = 0; cluster < num_clusters; cluster++)
                {
                    a_vp->elements[ cluster ] = initial_a_vp->elements[ cluster ];
                }

                ERE(ow_normalize_vector(a_vp, NORMALIZE_BY_SUM));
            }
            else
            {
                set_bug("Initial cluster priors vector is NULL.\n");
                return ERROR;
                
                /*
                warn_pso("Initial cluster priors vector is NULL: Using random priors.\n");
                for (i = 0; i < num_clusters; i++)
                {
                    a_vp->elements[i] = kjb_rand();
                }
                ERE(ow_add_scalar_to_vector(a_vp, 0.2 * kjb_rand() / num_clusters)); 
                ERE(ow_normalize_vector(a_vp, NORMALIZE_BY_SUM));
                */
            }

#ifdef DONT_USE_GAUSS_ROUTINE
            for (cluster = 0; cluster < num_clusters; cluster++)
            {
                temp = 0.0;

                if (result == ERROR) { NOTE_ERROR(); break; }

                block_diag_sizes_vp = block_diag_sizes_vvp->elements[ cluster ];

                num_block_diags = block_diag_sizes_vp->length;

                for (i = 0; i < num_block_diags; i++)
                {
                    block_diag_size = block_diag_sizes_vp->elements[i];

                    temp += block_diag_size * SAFE_LOG(sig_sqr);
                    temp += SAFE_LOG(1.0 + (block_diag_size * tau_sig_sqr_ratio));
                }

                log_sqrt_det_vp->elements[ cluster ] = temp / 2.0;
            }
            verify_vector(log_sqrt_det_vp);
#endif 
        }
        
        for (i = 0; i<num_points; i++)
        {
            if (result == ERROR) { NOTE_ERROR(); break; }

            result = get_matrix_row(&x_vp, feature_mp, i);
            if (result == ERROR) { NOTE_ERROR(); break; } 

            if ( (it == 0) && (fs_use_initialized_cluster_means_variances_and_priors != TRUE) )
            {
                /* E Step init */

                result = get_target_vector(&I_vp, num_clusters);
                if (result == ERROR) { NOTE_ERROR(); break; }
                
                for (cluster = 0; cluster < num_clusters; cluster++)
                {
                    double r = kjb_rand();
                    double p = pow(r, 5.0); /* Perhaps should be an option. */

                    if (result == ERROR) { NOTE_ERROR(); break; }

                    I_vp->elements[ cluster ] = p;

                    if (kjb_rand() < 1.0 / num_clusters) 
                    {
                        I_vp->elements[ cluster ] += 0.5; 
                    }
                }

                result = ow_scale_vector_by_sum(I_vp);
                if (result == ERROR) { NOTE_ERROR(); break; }

                result = ow_add_scalar_to_vector(I_vp, 0.2 * kjb_rand() / num_clusters); 
                if (result == ERROR) { NOTE_ERROR(); break; }

                result = ow_scale_vector_by_sum(I_vp);
                if (result == ERROR) { NOTE_ERROR(); break; }
            }
            
            if ( (it > 0) || (fs_use_initialized_cluster_means_variances_and_priors == TRUE) )
            {
                /* E Step */

                for (cluster = 0; cluster < num_clusters; cluster++)
                {
                    log_a = SAFE_LOG(a_vp->elements[ cluster ]);

#ifdef DONT_USE_GAUSS_ROUTINE                    
                    S_tilda_inv_mp = S_tilda_inv_mvp->elements[ cluster ];
                    
                    result = multiply_matrix_by_scalar(&S_inv_mp, S_tilda_inv_mp, (1.0 / sig_sqr));
                    if (result == ERROR) { NOTE_ERROR(); break; }
#else 
                    result = ERROR;
                    set_error("No Gauss routine for this case exists yet, sorry!");
                    if (result == ERROR) { NOTE_ERROR(); break; }

                    /*
                    if (result == ERROR) { NOTE_ERROR(); break; } 

                    result = get_matrix_row(&u_vp, u_mp, cluster);
                    if (result == ERROR) { NOTE_ERROR(); break; }

                    result = get_matrix_row(&var_vp, var_mp, cluster);
                    */
#endif 

#ifdef DONT_USE_GAUSS_ROUTINE

                    result = get_unity_vector(&ones_vp, num_features);
                    if (result == ERROR) { NOTE_ERROR(); break; }

                    result = multiply_vector_by_scalar(&u_vp, ones_vp, mu);
                    if (result == ERROR) { NOTE_ERROR(); break; }

                    result = subtract_vectors(&centered_x_vp, x_vp, u_vp);
                    if (result == ERROR) { NOTE_ERROR(); break; }

                    result = multiply_matrix_and_vector(&temp_vp, S_inv_mp, centered_x_vp);
                    if (result == ERROR) { NOTE_ERROR(); break; }

                    result = get_dot_product(centered_x_vp, temp_vp, &d);
                    if (result == ERROR) { NOTE_ERROR(); break; }

                    ASSERT_IS_NUMBER_DBL(d); 
                    ASSERT_IS_FINITE_DBL(d); 

                    I_vp->elements[ cluster ] = log_a - (0.5*d) - log_sqrt_det_vp->elements[ cluster ]; 
#else
                    result = ERROR;
                    set_error("No Gauss routine for this case exists yet, sorry!");
                    if (result == ERROR) { NOTE_ERROR(); break; }

                    /*
                    result = get_log_gaussian_density(x_vp, u_vp, var_vp, 
                                                      &(I_vp->elements[ cluster ]));
                    if (result == ERROR) { NOTE_ERROR(); break; }
                    I_vp->elements[ cluster ] += log_a; 
                    */
#endif 

                }

                if ( (held_out_indicator_vp != NULL) && (held_out_indicator_vp->elements[i] == 1) )
                {
                    held_out_log_likelihood += ow_exp_scale_by_sum_log_vector(I_vp);
                }
                else
                {
                    log_likelihood += ow_exp_scale_by_sum_log_vector(I_vp);
                }
            }

            ASSERT_IS_NEARLY_EQUAL_DBL(sum_vector_elements(I_vp), 1.0, 0.00001);

            if (result == ERROR) { NOTE_ERROR(); break; }

            if (P_mpp != NULL)
            {
                result = put_matrix_row(*P_mpp, I_vp, i);
            }

            /* M-Step */
            
            if ( (held_out_indicator_vp != NULL) && (held_out_indicator_vp->elements[i] == 1) )
            {
                /* Do not do the M-step for held out points */
            }
            
            else
            {
                for (cluster = 0; cluster < num_clusters; cluster++)
                {
                    if (result == ERROR) { NOTE_ERROR(); break; }

                    p = I_vp->elements[ cluster ];

                    S_tilda_inv_mp = S_tilda_inv_mvp->elements[ cluster ];

                    result = get_unity_vector(&ones_vp, num_features);
                    if (result == ERROR) { NOTE_ERROR(); break; }

                    /* Compute zero-order statistic */
                    result = multiply_matrix_and_vector(&temp_vp, S_tilda_inv_mp, ones_vp);
                    if (result == ERROR) { NOTE_ERROR(); break; }

                    result = get_dot_product(ones_vp, temp_vp, &d);
                    if (result == ERROR) { NOTE_ERROR(); break; }

                    zero_order_stat += (p * d);

                    /* Compute first-order statistic */
                    result = multiply_matrix_and_vector(&temp_vp, S_tilda_inv_mp, x_vp);
                    if (result == ERROR) { NOTE_ERROR(); break; }

                    result = get_dot_product(ones_vp, temp_vp, &d);
                    if (result == ERROR) { NOTE_ERROR(); break; }

                    first_order_stat += (p * d);

                    new_mu = first_order_stat;

                    /* Compute second-order statistic */
                    result = multiply_matrix_and_vector(&temp_vp, S_tilda_inv_mp, x_vp);
                    if (result == ERROR) { NOTE_ERROR(); break; }

                    result = get_dot_product(x_vp, temp_vp, &d);
                    if (result == ERROR) { NOTE_ERROR(); break; }

                    second_order_stat += (p * d);
                    
                    new_sig_sqr = second_order_stat;

                    p_sum_vp->elements[ cluster ] += p; 
                }
            }    
        }

        /* M-Step cleanup */

        ASSERT_IS_NOT_ZERO_DBL(zero_order_stat);
        mu = new_mu / zero_order_stat;

        sig_sqr = new_sig_sqr - (2.0 * mu * first_order_stat);
        sig_sqr = sig_sqr + (mu * mu * zero_order_stat);
        sig_sqr = sig_sqr / (sum_vector_elements(p_sum_vp));
        sig_sqr = sig_sqr / num_features;

#ifdef DONT_USE_GAUSS_ROUTINE
        for (cluster = 0; cluster < num_clusters; cluster++)
        {
            temp = 0.0;

            if (result == ERROR) { NOTE_ERROR(); break; }

            block_diag_sizes_vp = block_diag_sizes_vvp->elements[ cluster ];

            num_block_diags = block_diag_sizes_vp->length;

            for (i = 0; i < num_block_diags; i++)
            {
                block_diag_size = block_diag_sizes_vp->elements[i];

                temp += block_diag_size * SAFE_LOG(sig_sqr);
                temp += SAFE_LOG(1.0 + (block_diag_size * tau_sig_sqr_ratio));
            }

            log_sqrt_det_vp->elements[ cluster ] = temp / 2.0;
        }
        verify_vector(log_sqrt_det_vp);
#endif 

        result = scale_vector_by_sum(&a_vp, p_sum_vp); 
        if (result == ERROR) { NOTE_ERROR(); break; } 

        if (it > 0) 
        {
            diff = ABS_OF(log_likelihood - prev_log_likelihood);

            diff *= 2.0;
            diff /= (ABS_OF(log_likelihood) +  ABS_OF(prev_log_likelihood));

            verbose_pso(3, "%-3d: Log likelihood is %12e  |  %10e\n", 
                        it + 1, log_likelihood, diff); 

            held_out_diff = ABS_OF(held_out_log_likelihood - prev_held_out_log_likelihood);

            held_out_diff *= 2.0;
            held_out_diff /= (ABS_OF(held_out_log_likelihood) + ABS_OF(prev_held_out_log_likelihood));
            
            verbose_pso(3, "%-3d: Held out log likelihood is %12e  |  %10e\n",
                        it + 1, held_out_log_likelihood, held_out_diff); 

            if (fs_EM_stop_criterion_training_LL == TRUE)
            {
                if (diff < fs_iteration_tolerance) break; 
            }
            else if (fs_EM_stop_criterion_held_out_LL == TRUE)
            {
                if (held_out_diff < fs_iteration_tolerance) break;
            }

            prev_log_likelihood = log_likelihood;

            prev_held_out_log_likelihood = held_out_log_likelihood;
        }
    }

    if (result != ERROR)
    {
        if (log_likelihood_ptr != NULL)
        {
            *log_likelihood_ptr = log_likelihood;
        }

        if (held_out_log_likelihood_ptr != NULL)
        {
            *held_out_log_likelihood_ptr = held_out_log_likelihood;
        }

        if (num_iterations_ptr != NULL)
        {
            *num_iterations_ptr = it;
        }

        if (a_vpp != NULL) 
        {
            result = copy_vector(a_vpp, a_vp);
        }
    }
    
    if ((result != ERROR) && (mu_ptr != NULL))
    {
        *mu_ptr = mu;
    }

    if ((result != ERROR) && (sig_sqr_ptr != NULL))
    {
        *sig_sqr_ptr = sig_sqr;
    }

    free_vector(I_vp); 
    free_vector(p_sum_vp);
    free_vector(a_vp);
#ifdef DONT_USE_GAUSS_ROUTINE
    free_vector(log_sqrt_det_vp);
#endif 

    free_vector(x_vp); 
    free_vector(u_vp); 
    free_vector(centered_x_vp);
    free_vector(ones_vp);
    free_vector(temp_vp);

    free_matrix_vector(S_tilda_inv_mvp);
    free_matrix(S_inv_mp);

    if (result == ERROR) 
    {
        return ERROR; 
    }
    else 
    {
        return num_clusters;
    }
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *            get_GMM_blk_compound_sym_cov_1
 * 
 * This routine is the same as the get_GMM_blk_compound_sym_cov with the added
 * feature of estimating the tau_sig_sqr_ratio parameter automatically through
 * gradient ascent of the expected complete log likelihood in the M step.
 *
 * Finds a Gaussian mixture model (GMM) where the Gaussians have block compound
 * symmetrical covariances with shared parameters. Specifically, each feature
 * has the same mean (mu) and variance (sig^2 + tau^2). The covariance
 * between any pair of features is either (tau^2) or 0. This holds for all the
 * Gaussians in the mixture. However the block diagonal structure of the
 * covariance matrices of the different Gaussians in the mixture is different.
 * Here it is not assumed that the ratio (tau^2 / sig^2) is known beforehand
 * unlike the parent routine get_GMM_blk_compound_sym_cov.
 *
 * In particular, it fits:
 * |         p(x) = sum  a-sub-i *  g(mu-vec, cov-sub-i, x)
 * |                 i
 * where a-sub-i is the prior probability for the mixuture compoenent (cluster),
 * mu-vec is the mean vector with all components equal to mu, cov-sub-i is the
 * covariance matrix for the component i, and g(mu,cov,x) is a Gaussian with
 * mean mu and covariance cov. 
 *
 * The data matrix feature_mp is an N by M matrix where N is the number of data
 * points, and M is the number of features. 
 *
 * The argument block_diag_sizes_vvp specifies the block diagonal structures of
 * the covariances of the Gaussian components. The number of vectors in this
 * argument is equal to the number of mixture compoenents (clusters), K. Each
 * vector is a list of sizes of the block diagonals from top to bottom in the
 * corresponding covariance matrix. For eg., the vector corresponding to a
 * Gaussian component with all independent features would consist of 1 as all
 * its elements and the number of elements in the vector equal to M. And a
 * Gaussian component with two block diagonals of the same size in its
 * covariance matrix would be specified by a vector with two elements, each
 * equal to M/2.
 *
 * initial_a_vp, initial_mu, initial_sig_sqr and initial_tau_sig_sqr_ratio can
 * be used to specify the initial values of the parameters for EM.
 *
 * The model parameters are put into *a_vpp, *mu_ptr, *sig_sqr_ptr and
 * *tau_sig_sqr_ratio_ptr. Any of a_vpp, mu_ptr, sig_sqr_ptr or
 * tau_sig_sqr_ratio_ptr is NULL if that value is not needed.
 *
 * If P_mpp is not NULL, then the soft clustering (cluster membership) for each
 * data point is returned. In that case, *P_mpp will be N by K. 
 *
 * Returns :
 *    If the routine fails (due to storage allocation), then ERROR is returned
 *    with an error message being set. Otherwise NO_ERROR is returned. 
 *
 * Index: clustering, EM, GMM
 *
 * Note: The covariance structure assumed in this routine is often referred to
 * as "block compound-symmetry" structure especially in the mixed-models ANOVA
 * literature. It is useful in modeling data with repeated measures in ANOVA
 * using mixed-models. For example see:
 * http://www.asu.edu/sas/sasdoc/sashtml/stat/chap41/sect23.htm
 * http://www.tufts.edu/~gdallal/repeat2.htm 
 *
 * Authors: Prasad Gabbur, Kobus Barnard.
 * -----------------------------------------------------------------------------
*/
int get_GMM_blk_compound_sym_cov_1
(
    const Int_vector_vector* block_diag_sizes_vvp,
    const Matrix*            feature_mp,
    const Vector*            held_out_indicator_vp,
    const Vector*            initial_a_vp,
    const double             initial_mu,
    const double             initial_sig_sqr,
    const double             initial_tau_sig_sqr_ratio,
    Vector**                 a_vpp, 
    double*                  mu_ptr,
    double*                  sig_sqr_ptr,
    double*                  tau_sig_sqr_ratio_ptr,
    Matrix**                 P_mpp,
    double*                  log_likelihood_ptr,
    double*                  held_out_log_likelihood_ptr,
    int*                     num_iterations_ptr
)
{
    int num_clusters = block_diag_sizes_vvp->length;
    int num_points   = feature_mp->num_rows;
    int num_features = feature_mp->num_cols;
   
    Int_vector*  block_diag_sizes_vp = NULL;
    
    int      cluster;
    double   var_offset = fs_var_offset; 
    int      it, i, j, k, l, m;
    int      it_grad_ascent, it_M_step;
    double   mu, sig_sqr;
    double   new_mu, new_sig_sqr;
    double   prev_mu           = DBL_NOT_SET;
    double   prev_sig_sqr      = DBL_NOT_SET;
    double   diff_mu, diff_sig_sqr;
    double   tau_sig_sqr_ratio;
    double   new_tau_sig_sqr_ratio;
    double   prev_tau_sig_sqr_ratio = DBL_NOT_SET;
    double   diff_tau_sig_sqr_ratio;
    Vector*  I_vp          = NULL;
    Vector*  a_vp          = NULL;
    Vector*  p_sum_vp      = NULL;
    Vector*  x_vp          = NULL;
    Vector*  x_sub_vp      = NULL;
    Vector*  u_vp          = NULL;
    Vector*  centered_x_vp = NULL;
    Vector*  ones_vp       = NULL;
    Vector*  temp_vp       = NULL;
    double   log_likelihood               = DBL_HALF_MOST_NEGATIVE;
    double   prev_log_likelihood          = DBL_HALF_MOST_NEGATIVE;
    double   held_out_log_likelihood      = DBL_HALF_MOST_NEGATIVE;
    double   prev_held_out_log_likelihood = DBL_HALF_MOST_NEGATIVE;
    double   diff, held_out_diff;

    Matrix_vector* S_tilda_inv_mvp = NULL;
    Matrix*        S_tilda_inv_mp  = NULL;
    Matrix*        S_inv_mp        = NULL;
    Matrix*        temp_mp         = NULL;

    double zero_order_stat;
    double first_order_stat;
    double second_order_stat;

    double zero_order_stat_for_grad;
    double first_order_stat_for_grad;
    double second_order_stat_for_grad;
    double tau_sig_sqr_ratio_grad;
    double prev_tau_sig_sqr_ratio_grad;

    int     result = NO_ERROR;
#ifdef DONT_USE_GAUSS_ROUTINE
    Vector* log_sqrt_det_vp = NULL;
#endif 
    Matrix* I_1_mp = NULL;

#ifndef DONT_PLOT_GRAD_BUT_DO_GRAD_ASCENT    
    double  exp_comp_log_likelihood;
    double  pt_exp_comp_log_likelihood;

    Vector* plot_tau_sig_sqr_ratio_grad_vp  = NULL;
    Vector* plot_log_likelihood_vp          = NULL;
    Vector* log_p_vp                        = NULL;
    Vector* plot_exp_comp_log_likelihood_vp = NULL;
#endif

#ifdef DONT_USE_GAUSS_ROUTINE
    double temp;
#endif
    double p, d;
    double log_a;
    double inv_S_blk_grad;
    int    num_block_diags;
    int    block_diag_size;
    int    sum_block_sizes;

    double rate_grad_ascent = 1.0;
    double prev_abs_change  = 1.0;
    int    update_mu_sig_sqr;

    if (P_mpp != NULL)
    {
        ERE(get_target_matrix(P_mpp, num_points, num_clusters));
    }

    ERE(get_target_matrix(&I_1_mp, num_points, num_clusters));

#ifndef DONT_PLOT_GRAD_BUT_DO_GRAD_ASCENT    
    ERE(get_target_vector(&plot_tau_sig_sqr_ratio_grad_vp, 1000/0.01));
#endif

    if (held_out_indicator_vp != NULL)
    {
        if (held_out_indicator_vp->length != num_points)
        {
            result = ERROR;
        }
    }
            
    if (    (get_target_vector(&I_vp, num_clusters) == ERROR)
         || (get_target_vector(&a_vp, num_clusters) == ERROR) 
#ifdef DONT_USE_GAUSS_ROUTINE
         || (get_target_vector(&log_sqrt_det_vp, num_clusters) == ERROR) 
#endif 
       )
    {
        result = ERROR;
    }

    dbi(num_clusters); 

    if (result != ERROR)
    {
        result = get_target_matrix_vector(&S_tilda_inv_mvp, num_clusters);
    }

    tau_sig_sqr_ratio = initial_tau_sig_sqr_ratio;

    for (cluster = 0; cluster < num_clusters; cluster++)
    {
        if (result == ERROR) { NOTE_ERROR(); break; }

        sum_block_sizes = 0;

        /*
        S_tilda_inv_mp = S_tilda_inv_mvp->elements[ cluster ];

        result = get_zero_matrix(&S_tilda_inv_mp, num_features, num_features);
        if (result == ERROR) { NOTE_ERROR(); break; }
        */

        result = get_zero_matrix(&(S_tilda_inv_mvp->elements[ cluster ]), num_features, num_features);
        if (result == ERROR) { NOTE_ERROR(); break; }
        
        S_tilda_inv_mp = S_tilda_inv_mvp->elements[ cluster ];

        block_diag_sizes_vp = block_diag_sizes_vvp->elements[ cluster ];

        db_irv(block_diag_sizes_vp);

        num_block_diags = block_diag_sizes_vp->length;

        for (i = 0, j = 0, k = 0; i < num_block_diags; i++)
        {
            block_diag_size = block_diag_sizes_vp->elements[i];

            sum_block_sizes = sum_block_sizes + block_diag_size;

            l = j;

            for (j = l; j < (l + block_diag_size); j++)
            {
                for (k = l; k < (l + block_diag_size); k++)
                {
                    if (j == k)
                    {
                        S_tilda_inv_mp->elements[j][k] = 1 + ((block_diag_size - 1) * tau_sig_sqr_ratio);
                        S_tilda_inv_mp->elements[j][k] = S_tilda_inv_mp->elements[j][k] / \
                                                         (1 + (block_diag_size * tau_sig_sqr_ratio));
                    }
                    else
                    {
                        S_tilda_inv_mp->elements[j][k] = -tau_sig_sqr_ratio / \
                                                         (1 + (block_diag_size * tau_sig_sqr_ratio));
                    }
                }
            }
        }

        if (sum_block_sizes != num_features)
        {
            set_bug("Block diagonal sizes for cluster %d do not sum \
                    to the feature dimensionality:", (cluster + 1));
            return ERROR;
        }
    }

    for (it = 0; it < fs_max_num_iterations; it++)
    {
#ifndef DONT_PLOT_GRAD_BUT_DO_GRAD_ASCENT

        ERE(get_target_vector(&log_p_vp, num_clusters));
        ERE(get_target_vector(&plot_log_likelihood_vp, 1000/0.01));
        ERE(get_target_vector(&plot_exp_comp_log_likelihood_vp, 1000/0.01));

        it_grad_ascent   = 0;
        tau_sig_sqr_ratio = initial_tau_sig_sqr_ratio;

        do
        {
        dbi(it_grad_ascent);

        exp_comp_log_likelihood = 0.0;
#endif
        log_likelihood = 0.0;

        held_out_log_likelihood = 0.0;

        if (result == ERROR) { NOTE_ERROR(); break; }

        result = get_zero_vector(&p_sum_vp, num_clusters);
        if (result == ERROR) { NOTE_ERROR(); break; }

        if ( (it == 0) && (fs_use_initialized_cluster_means_variances_and_priors == TRUE) )
        {
            mu                = initial_mu;
            sig_sqr           = initial_sig_sqr + var_offset;
#ifdef DONT_PLOT_GRAD_BUT_DO_GRAD_ASCENT           
            tau_sig_sqr_ratio = initial_tau_sig_sqr_ratio;
#endif
            prev_mu                = mu;
            prev_sig_sqr           = sig_sqr;
            prev_tau_sig_sqr_ratio = tau_sig_sqr_ratio;

            if (initial_a_vp != NULL)
            {
                ERE(get_target_vector(&a_vp, num_clusters));

                for (cluster = 0; cluster < num_clusters; cluster++)
                {
                    a_vp->elements[ cluster ] = initial_a_vp->elements[ cluster ];
                }

                ERE(ow_normalize_vector(a_vp, NORMALIZE_BY_SUM));
            }
            else
            {
                set_bug("Initial cluster priors vector is NULL.\n");
                return ERROR;
                
                /*
                warn_pso("Initial cluster priors vector is NULL: Using random priors.\n");
                for (i = 0; i < num_clusters; i++)
                {
                    a_vp->elements[i] = kjb_rand();
                }
                ERE(ow_add_scalar_to_vector(a_vp, 0.2 * kjb_rand() / num_clusters)); 
                ERE(ow_normalize_vector(a_vp, NORMALIZE_BY_SUM));
                */
            }

#ifdef DONT_USE_GAUSS_ROUTINE
            for (cluster = 0; cluster < num_clusters; cluster++)
            {
                temp = 0.0;

                if (result == ERROR) { NOTE_ERROR(); break; }

                block_diag_sizes_vp = block_diag_sizes_vvp->elements[ cluster ];

                num_block_diags = block_diag_sizes_vp->length;

                for (i = 0; i < num_block_diags; i++)
                {
                    block_diag_size = block_diag_sizes_vp->elements[i];

                    temp += block_diag_size * SAFE_LOG(sig_sqr);
                    temp += SAFE_LOG(1.0 + (block_diag_size * tau_sig_sqr_ratio));
                }

                log_sqrt_det_vp->elements[ cluster ] = temp / 2.0;
            }
            verify_vector(log_sqrt_det_vp);
#endif 
        }
        
        /* E Step */
        for (i = 0; i<num_points; i++)
        {
            if (result == ERROR) { NOTE_ERROR(); break; }

            result = get_matrix_row(&x_vp, feature_mp, i);
            if (result == ERROR) { NOTE_ERROR(); break; } 

            if ( (it == 0) && (fs_use_initialized_cluster_means_variances_and_priors != TRUE) )
            {
                /* E Step init */

                result = get_target_vector(&I_vp, num_clusters);
                if (result == ERROR) { NOTE_ERROR(); break; }
                
                for (cluster = 0; cluster < num_clusters; cluster++)
                {
                    double r = kjb_rand();
                    double p = pow(r, 5.0); /* Perhaps should be an option. */

                    if (result == ERROR) { NOTE_ERROR(); break; }

                    I_vp->elements[ cluster ] = p;

                    if (kjb_rand() < 1.0 / num_clusters) 
                    {
                        I_vp->elements[ cluster ] += 0.5; 
                    }
                }

                result = ow_scale_vector_by_sum(I_vp);
                if (result == ERROR) { NOTE_ERROR(); break; }

                result = ow_add_scalar_to_vector(I_vp, 0.2 * kjb_rand() / num_clusters); 
                if (result == ERROR) { NOTE_ERROR(); break; }

                result = ow_scale_vector_by_sum(I_vp);
                if (result == ERROR) { NOTE_ERROR(); break; }
            }
            
            if ( (it > 0) || (fs_use_initialized_cluster_means_variances_and_priors == TRUE) )
            {
                /* E Step */

                for (cluster = 0; cluster < num_clusters; cluster++)
                {
                    log_a = SAFE_LOG(a_vp->elements[ cluster ]);

#ifdef DONT_USE_GAUSS_ROUTINE                    
                    S_tilda_inv_mp = S_tilda_inv_mvp->elements[ cluster ];
                    
                    result = multiply_matrix_by_scalar(&S_inv_mp, S_tilda_inv_mp, (1.0 / sig_sqr));
                    if (result == ERROR) { NOTE_ERROR(); break; }
#else 
                    result = ERROR;
                    set_error("No Gauss routine for this case exists yet, sorry!");
                    if (result == ERROR) { NOTE_ERROR(); break; }

                    /*
                    if (result == ERROR) { NOTE_ERROR(); break; } 

                    result = get_matrix_row(&u_vp, u_mp, cluster);
                    if (result == ERROR) { NOTE_ERROR(); break; }

                    result = get_matrix_row(&var_vp, var_mp, cluster);
                    */
#endif 

#ifdef DONT_USE_GAUSS_ROUTINE

                    result = get_unity_vector(&ones_vp, num_features);
                    if (result == ERROR) { NOTE_ERROR(); break; }

                    result = multiply_vector_by_scalar(&u_vp, ones_vp, mu);
                    if (result == ERROR) { NOTE_ERROR(); break; }

                    result = subtract_vectors(&centered_x_vp, x_vp, u_vp);
                    if (result == ERROR) { NOTE_ERROR(); break; }

                    result = multiply_matrix_and_vector(&temp_vp, S_inv_mp, centered_x_vp);
                    if (result == ERROR) { NOTE_ERROR(); break; }

                    result = get_dot_product(centered_x_vp, temp_vp, &d);
                    if (result == ERROR) { NOTE_ERROR(); break; }

                    ASSERT_IS_NUMBER_DBL(d); 
                    ASSERT_IS_FINITE_DBL(d); 

                    I_vp->elements[ cluster ] = log_a - (0.5*d) - log_sqrt_det_vp->elements[ cluster ]; 
#else
                    result = ERROR;
                    set_error("No Gauss routine for this case exists yet, sorry!");
                    if (result == ERROR) { NOTE_ERROR(); break; }

                    /*
                    result = get_log_gaussian_density(x_vp, u_vp, var_vp, 
                                                      &(I_vp->elements[ cluster ]));
                    if (result == ERROR) { NOTE_ERROR(); break; }
                    I_vp->elements[ cluster ] += log_a; 
                    */
#endif 

#ifndef DONT_PLOT_GRAD_BUT_DO_GRAD_ASCENT
                    log_p_vp->elements[ cluster ] =  - (0.5*d) - log_sqrt_det_vp->elements[ cluster ];
#endif
                }

                if ( (held_out_indicator_vp != NULL) && (held_out_indicator_vp->elements[i] == 1) )
                {
                    held_out_log_likelihood += ow_exp_scale_by_sum_log_vector(I_vp);
                }
                else
                {
                    log_likelihood += ow_exp_scale_by_sum_log_vector(I_vp);
                }

#ifndef DONT_PLOT_GRAD_BUT_DO_GRAD_ASCENT
                ERE(get_dot_product(log_p_vp, I_vp, &pt_exp_comp_log_likelihood));
                exp_comp_log_likelihood += pt_exp_comp_log_likelihood; 
#endif

            }

            ASSERT_IS_NEARLY_EQUAL_DBL(sum_vector_elements(I_vp), 1.0, 0.00001);

            if (result == ERROR) { NOTE_ERROR(); break; }

            if (P_mpp != NULL)
            {
                result = put_matrix_row(*P_mpp, I_vp, i);
            }

            if (result == ERROR) { NOTE_ERROR(); break; }

            result = put_matrix_row(I_1_mp, I_vp, i);
            if (result == ERROR) { NOTE_ERROR(); break; }
        }

#ifndef DONT_PLOT_GRAD_BUT_DO_GRAD_ASCENT

        verbose_pso(3, "%-3d: tau_sig_sqr_ratio is %12e  |  %10e\n",
                            it_grad_ascent + 1, tau_sig_sqr_ratio, diff_tau_sig_sqr_ratio);

        plot_log_likelihood_vp->elements[it_grad_ascent]          = log_likelihood;
        plot_exp_comp_log_likelihood_vp->elements[it_grad_ascent] = exp_comp_log_likelihood;

        tau_sig_sqr_ratio                                 = tau_sig_sqr_ratio + 0.01;

        it_grad_ascent = it_grad_ascent + 1;
        
        /* Form S_tilda_mvp from the new value of tau_sig_sqr_ratio */
        for (cluster = 0; cluster < num_clusters; cluster++)
        {
            if (result == ERROR) { NOTE_ERROR(); break; }

            result = get_zero_matrix(&(S_tilda_inv_mvp->elements[ cluster ]), num_features, num_features);
            if (result == ERROR) { NOTE_ERROR(); break; }

            S_tilda_inv_mp = S_tilda_inv_mvp->elements[ cluster ];

            block_diag_sizes_vp = block_diag_sizes_vvp->elements[ cluster ];

            num_block_diags = block_diag_sizes_vp->length;

            for (i = 0, j = 0, k = 0; i < num_block_diags; i++)
            {
                block_diag_size = block_diag_sizes_vp->elements[i];

                l = j;

                for (j = l; j < (l + block_diag_size); j++)
                {
                    for (k = l; k < (l + block_diag_size); k++)
                    {
                        if (j == k)
                        {
                            S_tilda_inv_mp->elements[j][k] = 1 + ((block_diag_size - 1) * tau_sig_sqr_ratio);
                            S_tilda_inv_mp->elements[j][k] = S_tilda_inv_mp->elements[j][k] / \
                                                             (1 + (block_diag_size * tau_sig_sqr_ratio));
                        }
                        else
                        {
                            S_tilda_inv_mp->elements[j][k] = -tau_sig_sqr_ratio / \
                                                             (1 + (block_diag_size * tau_sig_sqr_ratio));
                        }
                    }
                }
            }
        }

        } while (tau_sig_sqr_ratio < 1000);

        ERE(write_col_vector(plot_log_likelihood_vp, "log_likelihood_vp"));
        ERE(write_col_vector(plot_log_likelihood_vp, "exp_comp_log_likelihood_vp"));

        tau_sig_sqr_ratio = initial_tau_sig_sqr_ratio;

        /* Form S_tilda_mvp from the new value of tau_sig_sqr_ratio */
        for (cluster = 0; cluster < num_clusters; cluster++)
        {
            if (result == ERROR) { NOTE_ERROR(); break; }

            result = get_zero_matrix(&(S_tilda_inv_mvp->elements[ cluster ]), num_features, num_features);
            if (result == ERROR) { NOTE_ERROR(); break; }

            S_tilda_inv_mp = S_tilda_inv_mvp->elements[ cluster ];

            block_diag_sizes_vp = block_diag_sizes_vvp->elements[ cluster ];

            num_block_diags = block_diag_sizes_vp->length;

            for (i = 0, j = 0, k = 0; i < num_block_diags; i++)
            {
                block_diag_size = block_diag_sizes_vp->elements[i];

                l = j;

                for (j = l; j < (l + block_diag_size); j++)
                {
                    for (k = l; k < (l + block_diag_size); k++)
                    {
                        if (j == k)
                        {
                            S_tilda_inv_mp->elements[j][k] = 1 + ((block_diag_size - 1) * tau_sig_sqr_ratio);
                            S_tilda_inv_mp->elements[j][k] = S_tilda_inv_mp->elements[j][k] / \
                                                             (1 + (block_diag_size * tau_sig_sqr_ratio));
                        }
                        else
                        {
                            S_tilda_inv_mp->elements[j][k] = -tau_sig_sqr_ratio / \
                                                             (1 + (block_diag_size * tau_sig_sqr_ratio));
                        }
                    }
                }
            }
        }

#endif
        
        /* M-Step */
#ifdef DONT_PLOT_GRAD_BUT_DO_GRAD_ASCENT
        update_mu_sig_sqr = TRUE;
#else
        dbf(mu);
        dbf(sig_sqr);
        update_mu_sig_sqr = FALSE;
#endif

        it_M_step = 0;

        do
        {
            it_grad_ascent              = 0;
            prev_tau_sig_sqr_ratio_grad = DBL_NOT_SET;
            rate_grad_ascent            = 1.0;

            do /*for (it_grad_ascent = 0; it_grad_ascent < 1000; it_grad_ascent++)*/
            {
                new_mu            = 0.0;
                new_sig_sqr       = 0.0;
                zero_order_stat   = 0.0;
                first_order_stat  = 0.0;
                second_order_stat = 0.0;

                tau_sig_sqr_ratio_grad     = 0.0;
                zero_order_stat_for_grad   = 0.0;
                first_order_stat_for_grad  = 0.0;
                second_order_stat_for_grad = 0.0;

                result = get_zero_vector(&p_sum_vp, num_clusters);
                if (result == ERROR) { NOTE_ERROR(); break; }

                for (i = 0; i<num_points; i++)
                {     
                    if ( (held_out_indicator_vp != NULL) && (held_out_indicator_vp->elements[i] == 1) )
                    {
                        /* Do not do the M-step for held out points */
                    } 
                    else
                    {
                        if (result == ERROR) { NOTE_ERROR(); break; } 

                        result = get_matrix_row(&x_vp, feature_mp, i);
                        if (result == ERROR) { NOTE_ERROR(); break; } 

                        result = get_matrix_row(&I_vp, I_1_mp, i);
                        if (result == ERROR) { NOTE_ERROR(); break; }

                        for (cluster = 0; cluster < num_clusters; cluster++)
                        {
                            if (result == ERROR) { NOTE_ERROR(); break; }

                            p = I_vp->elements[ cluster ];

                            S_tilda_inv_mp = S_tilda_inv_mvp->elements[ cluster ];

                            /* Compute a gradient term */
                            block_diag_sizes_vp = block_diag_sizes_vvp->elements[ cluster ];

                            num_block_diags = block_diag_sizes_vp->length;

                            for (j = 0; j < num_block_diags; j++)
                            {
                                block_diag_size = block_diag_sizes_vp->elements[j];

                                tau_sig_sqr_ratio_grad -= (0.5 * p * block_diag_size) /\
                                                               (1.0 + (block_diag_size * tau_sig_sqr_ratio));
                            }

                            /* Compute zero-order statistics */
                            result = get_unity_vector(&ones_vp, num_features);
                            if (result == ERROR) { NOTE_ERROR(); break; }

                            result = multiply_matrix_and_vector(&temp_vp, S_tilda_inv_mp, ones_vp);
                            if (result == ERROR) { NOTE_ERROR(); break; }

                            result = get_dot_product(ones_vp, temp_vp, &d);
                            if (result == ERROR) { NOTE_ERROR(); break; }

                            zero_order_stat += (p * d);

                            block_diag_sizes_vp = block_diag_sizes_vvp->elements[ cluster ];

                            num_block_diags = block_diag_sizes_vp->length;

                            for (j = 0; j < num_block_diags; j++)
                            {
                                block_diag_size = block_diag_sizes_vp->elements[j];

                                inv_S_blk_grad  = 1.0/(1.0 +  (block_diag_size*tau_sig_sqr_ratio));
                                inv_S_blk_grad  = inv_S_blk_grad*inv_S_blk_grad; 
                                result = get_initialized_matrix(&temp_mp, block_diag_size, block_diag_size, inv_S_blk_grad);
                                if (result == ERROR) { NOTE_ERROR(); break; }

                                result = get_unity_vector(&ones_vp, block_diag_size);
                                if (result == ERROR) { NOTE_ERROR(); break; }

                                result = multiply_matrix_and_vector(&temp_vp, temp_mp, ones_vp);                    
                                if (result == ERROR) { NOTE_ERROR(); break; }

                                result = get_dot_product(ones_vp, temp_vp, &d);
                                if (result == ERROR) { NOTE_ERROR(); break; }

                                zero_order_stat_for_grad += (p * d);
                            }

                            /* Compute first-order statistics */
                            result = multiply_matrix_and_vector(&temp_vp, S_tilda_inv_mp, x_vp);                    
                            if (result == ERROR) { NOTE_ERROR(); break; }

                            result = get_unity_vector(&ones_vp, num_features);
                            if (result == ERROR) { NOTE_ERROR(); break; }

                            result = get_dot_product(ones_vp, temp_vp, &d);
                            if (result == ERROR) { NOTE_ERROR(); break; }

                            first_order_stat += (p * d);

                            new_mu = first_order_stat;

                            block_diag_sizes_vp = block_diag_sizes_vvp->elements[ cluster ];

                            num_block_diags = block_diag_sizes_vp->length;

                            for (j = 0, k = 0; j < num_block_diags; j++)
                            {
                                block_diag_size = block_diag_sizes_vp->elements[j];

                                result = get_target_vector(&x_sub_vp, block_diag_size);
                                if (result == ERROR) { NOTE_ERROR(); break; }

                                l = k;

                                for (k = l, m = 0; k < (l + block_diag_size); k++, m++)
                                {
                                    x_sub_vp->elements[m] = x_vp->elements[k];
                                }

                                inv_S_blk_grad  = 1.0/(1.0 +  (block_diag_size*tau_sig_sqr_ratio));
                                inv_S_blk_grad  = inv_S_blk_grad*inv_S_blk_grad; 
                                result = get_initialized_matrix(&temp_mp, block_diag_size, block_diag_size, inv_S_blk_grad);
                                if (result == ERROR) { NOTE_ERROR(); break; }

                                result = multiply_matrix_and_vector(&temp_vp, temp_mp, x_sub_vp);                    
                                if (result == ERROR) { NOTE_ERROR(); break; }

                                result = get_unity_vector(&ones_vp, block_diag_size);
                                if (result == ERROR) { NOTE_ERROR(); break; }

                                result = get_dot_product(ones_vp, temp_vp, &d);
                                if (result == ERROR) { NOTE_ERROR(); break; }

                                first_order_stat_for_grad += (p * d);
                            }

                            /* Compute second-order statistics */
                            result = multiply_matrix_and_vector(&temp_vp, S_tilda_inv_mp, x_vp);
                            if (result == ERROR) { NOTE_ERROR(); break; }

                            result = get_dot_product(x_vp, temp_vp, &d);
                            if (result == ERROR) { NOTE_ERROR(); break; }

                            second_order_stat += (p * d);

                            new_sig_sqr = second_order_stat;

                            block_diag_sizes_vp = block_diag_sizes_vvp->elements[ cluster ];

                            num_block_diags = block_diag_sizes_vp->length;

                            for (j = 0, k = 0; j < num_block_diags; j++)
                            {
                                block_diag_size = block_diag_sizes_vp->elements[j];

                                result = get_target_vector(&x_sub_vp, block_diag_size);
                                if (result == ERROR) { NOTE_ERROR(); break; }

                                l = k;

                                for (k = l, m = 0; k < (l + block_diag_size); k++, m++)
                                {
                                    x_sub_vp->elements[m] = x_vp->elements[k];
                                }

                                inv_S_blk_grad = (1.0/(1.0 +  (block_diag_size*tau_sig_sqr_ratio)));
                                inv_S_blk_grad = inv_S_blk_grad*inv_S_blk_grad; 
                                result = get_initialized_matrix(&temp_mp, block_diag_size, block_diag_size, inv_S_blk_grad);
                                if (result == ERROR) { NOTE_ERROR(); break; }

                                result = multiply_matrix_and_vector(&temp_vp, temp_mp, x_sub_vp);                    
                                if (result == ERROR) { NOTE_ERROR(); break; }

                                result = get_dot_product(x_sub_vp, temp_vp, &d);
                                if (result == ERROR) { NOTE_ERROR(); break; }

                                second_order_stat_for_grad += (p * d);
                            }

                            p_sum_vp->elements[ cluster ] += p; 
                        }
                    }    
                }

                /* M-Step cleanup */
                if (update_mu_sig_sqr == TRUE)
                {
                    ASSERT_IS_NOT_ZERO_DBL(zero_order_stat);
                    mu = new_mu / zero_order_stat;

                    sig_sqr = new_sig_sqr - (2.0 * mu * first_order_stat);
                    sig_sqr = sig_sqr + (mu * mu * zero_order_stat);
                    sig_sqr = sig_sqr / (sum_vector_elements(p_sum_vp));
                    sig_sqr = sig_sqr / num_features;
            
                    diff_mu      = ABS_OF(prev_mu - mu);
                    diff_sig_sqr = ABS_OF(prev_sig_sqr - sig_sqr); 

                    verbose_pso(3, "%-3d, %-3d, %-3d: mu is %12e  |  %10e\n",
                                it + 1, it_M_step + 1, it_grad_ascent + 1, mu, diff_mu);
            
                    verbose_pso(3, "%-3d, %-3d, %-3d: sig_sqr is %12e  |  %10e\n",
                                it + 1, it_M_step + 1, it_grad_ascent + 1, sig_sqr, diff_sig_sqr); 
                }

                ASSERT_IS_NOT_ZERO_DBL(sig_sqr);
                tau_sig_sqr_ratio_grad += (mu * mu * zero_order_stat_for_grad) / (2.0 * sig_sqr);
                tau_sig_sqr_ratio_grad -= (2.0 * mu * first_order_stat_for_grad) / (2.0 * sig_sqr);
                tau_sig_sqr_ratio_grad += second_order_stat_for_grad / (2.0 * sig_sqr);

                verbose_pso(3, "%-3d, %-3d, %-3d: tau_sig_sqr_ratio_grad is %12e\n", 
                            it + 1, it_M_step + 1, it_grad_ascent + 1, tau_sig_sqr_ratio_grad);

#ifdef DONT_PLOT_GRAD_BUT_DO_GRAD_ASCENT                
                /* tau_sig_sqr_ratio gets updated here using gradient ascent */
                new_tau_sig_sqr_ratio = tau_sig_sqr_ratio + \
                                        (rate_grad_ascent * SIGN_OF(tau_sig_sqr_ratio_grad));

                verbose_pso(3, "%-3d, %-3d, %-3d: new_tau_sig_sqr_ratio is %12e\n", 
                            it + 1, it_M_step + 1, it_grad_ascent + 1, new_tau_sig_sqr_ratio);

                /* The following is a hack to avoid going below zero for the
                 * tau_sig_sqr_ratio parameter. When we get close to zero,
                 * restart from the initial value(?) for this parameter. Also
                 * reduce the rate of gradient ascent in order to explore the
                 * space in between. The assumption is that the likelihood
                 * maxima lies somewhere between zero and the initial value. */
                if (new_tau_sig_sqr_ratio <= DBL_EPSILON)
                {
                    /*
                    tau_sig_sqr_ratio = initial_tau_sig_sqr_ratio;
                    */
                    rate_grad_ascent  = rate_grad_ascent / 2.0;

                    dbe(rate_grad_ascent);

                    /*
                    if ( ABS_OF(rate_grad_ascent * tau_sig_sqr_ratio_grad) >= 0.00001 )
                    */
                    if ( rate_grad_ascent >= 0.00001 )
                    {
                        continue;
                    }
                }
                else
                {
                    tau_sig_sqr_ratio = new_tau_sig_sqr_ratio;
                }

                diff_tau_sig_sqr_ratio = ABS_OF(tau_sig_sqr_ratio - prev_tau_sig_sqr_ratio);
                prev_tau_sig_sqr_ratio = tau_sig_sqr_ratio;
                prev_abs_change = diff_tau_sig_sqr_ratio;

                /* Mocking simulated annealing by reducing the rate of gradient
                 * ascent when two successive gradients are of the opposite
                 * sign. */
                if ( prev_tau_sig_sqr_ratio_grad != DBL_NOT_SET )
                {
                    if (    (prev_tau_sig_sqr_ratio_grad <= 0 && tau_sig_sqr_ratio_grad > 0) 
                         || (prev_tau_sig_sqr_ratio_grad >= 0 && tau_sig_sqr_ratio_grad < 0) )
                    {
                        rate_grad_ascent = rate_grad_ascent / 2.0;

                        dbe(rate_grad_ascent);
                    }
                }
                prev_tau_sig_sqr_ratio_grad = tau_sig_sqr_ratio_grad;
#else
                plot_tau_sig_sqr_ratio_grad_vp->elements[it_grad_ascent] = tau_sig_sqr_ratio_grad;
                tau_sig_sqr_ratio                                        = tau_sig_sqr_ratio + 0.01;
#endif
                verbose_pso(3, "%-3d, %-3d, %-3d: tau_sig_sqr_ratio is %12e  |  %10e\n",
                            it + 1, it_M_step + 1, it_grad_ascent + 1, tau_sig_sqr_ratio, diff_tau_sig_sqr_ratio);

                /* Form S_tilda_mvp from the new value of tau_sig_sqr_ratio */
                for (cluster = 0; cluster < num_clusters; cluster++)
                {
                    if (result == ERROR) { NOTE_ERROR(); break; }

                    result = get_zero_matrix(&(S_tilda_inv_mvp->elements[ cluster ]), num_features, num_features);
                    if (result == ERROR) { NOTE_ERROR(); break; }

                    S_tilda_inv_mp = S_tilda_inv_mvp->elements[ cluster ];

                    block_diag_sizes_vp = block_diag_sizes_vvp->elements[ cluster ];

                    num_block_diags = block_diag_sizes_vp->length;

                    for (i = 0, j = 0, k = 0; i < num_block_diags; i++)
                    {
                        block_diag_size = block_diag_sizes_vp->elements[i];

                        l = j;

                        for (j = l; j < (l + block_diag_size); j++)
                        {
                            for (k = l; k < (l + block_diag_size); k++)
                            {
                                if (j == k)
                                {
                                    S_tilda_inv_mp->elements[j][k] = 1 + ((block_diag_size - 1) * tau_sig_sqr_ratio);
                                    S_tilda_inv_mp->elements[j][k] = S_tilda_inv_mp->elements[j][k] / \
                                                                     (1 + (block_diag_size * tau_sig_sqr_ratio));
                                }
                                else
                                {
                                    S_tilda_inv_mp->elements[j][k] = -tau_sig_sqr_ratio / \
                                                                     (1 + (block_diag_size * tau_sig_sqr_ratio));
                                }
                            }
                        }
                    }
                }

#ifdef DONT_USE_GAUSS_ROUTINE
                for (cluster = 0; cluster < num_clusters; cluster++)
                {
                    temp = 0.0;

                    if (result == ERROR) { NOTE_ERROR(); break; }

                    block_diag_sizes_vp = block_diag_sizes_vvp->elements[ cluster ];

                    num_block_diags = block_diag_sizes_vp->length;

                    for (i = 0; i < num_block_diags; i++)
                    {
                        block_diag_size = block_diag_sizes_vp->elements[i];

                        temp += block_diag_size * SAFE_LOG(sig_sqr);
                        temp += SAFE_LOG(1.0 + (block_diag_size * tau_sig_sqr_ratio));
                    }

                    log_sqrt_det_vp->elements[ cluster ] = temp / 2.0;
                }
                verify_vector(log_sqrt_det_vp);
#endif 

                result = scale_vector_by_sum(&a_vp, p_sum_vp); 
                if (result == ERROR) { NOTE_ERROR(); break; } 

#ifdef DONT_PLOT_GRAD_BUT_DO_GRAD_ASCENT                     
                if ( diff_tau_sig_sqr_ratio < 0.00001 )
                {
                    update_mu_sig_sqr = TRUE;
                    break;
                }
#endif                
                update_mu_sig_sqr = FALSE;

                it_grad_ascent = it_grad_ascent + 1;

            }
#ifdef DONT_PLOT_GRAD_BUT_DO_GRAD_ASCENT     
            while (1); 
#else
            while (tau_sig_sqr_ratio < 1000);
#endif

            if (update_mu_sig_sqr == TRUE)
            {
                if (    diff_mu      < 0.00001 
                     && diff_sig_sqr < 0.00001 
                     && prev_mu      != DBL_NOT_SET
                     && prev_sig_sqr != DBL_NOT_SET )
                {
                    update_mu_sig_sqr = FALSE;
                }
            }

            prev_mu      = mu;
            prev_sig_sqr = sig_sqr;

            it_M_step = it_M_step + 1;

        } while (update_mu_sig_sqr == TRUE);

#ifndef DONT_PLOT_GRAD_BUT_DO_GRAD_ASCENT
        /*
        int plot_id = plot_open();
        ERE(plot_vector(plot_id, plot_tau_sig_sqr_ratio_grad_vp, 0.0, 0.1, NULL));
        */
        ERE(write_col_vector(plot_tau_sig_sqr_ratio_grad_vp, "tau_sig_sqr_ratio_grad_vp"));
        break;
#endif
        
        if (it > 0) 
        {
            diff = ABS_OF(log_likelihood - prev_log_likelihood);

            diff *= 2.0;
            diff /= (ABS_OF(log_likelihood) +  ABS_OF(prev_log_likelihood));

            verbose_pso(3, "%-3d: Log likelihood is %12e  |  %10e\n", 
                        it + 1, log_likelihood, diff); 

            held_out_diff = ABS_OF(held_out_log_likelihood - prev_held_out_log_likelihood);

            held_out_diff *= 2.0;
            held_out_diff /= (ABS_OF(held_out_log_likelihood) + ABS_OF(prev_held_out_log_likelihood));
            
            verbose_pso(3, "%-3d: Held out log likelihood is %12e  |  %10e\n",
                        it + 1, held_out_log_likelihood, held_out_diff); 

            if (fs_EM_stop_criterion_training_LL == TRUE)
            {
                if (diff < fs_iteration_tolerance) break; 
            }
            else if (fs_EM_stop_criterion_held_out_LL == TRUE)
            {
                if (held_out_diff < fs_iteration_tolerance) break;
            }

            prev_log_likelihood = log_likelihood;

            prev_held_out_log_likelihood = held_out_log_likelihood;
        }
    }

    if (result != ERROR)
    {
        if (log_likelihood_ptr != NULL)
        {
            *log_likelihood_ptr = log_likelihood;
        }

        if (held_out_log_likelihood_ptr != NULL)
        {
            *held_out_log_likelihood_ptr = held_out_log_likelihood;
        }

        if (num_iterations_ptr != NULL)
        {
            *num_iterations_ptr = it;
        }

        if (a_vpp != NULL) 
        {
            result = copy_vector(a_vpp, a_vp);
        }
    }
    
    if ((result != ERROR) && (mu_ptr != NULL))
    {
        *mu_ptr = mu;
    }

    if ((result != ERROR) && (sig_sqr_ptr != NULL))
    {
        *sig_sqr_ptr = sig_sqr;
    }

    if ((result != ERROR) && (tau_sig_sqr_ratio_ptr != NULL))
    {
        *tau_sig_sqr_ratio_ptr = tau_sig_sqr_ratio;
    }
    
    free_vector(I_vp); 
    free_vector(p_sum_vp);
    free_vector(a_vp);
#ifdef DONT_USE_GAUSS_ROUTINE
    free_vector(log_sqrt_det_vp);
#endif 

    free_vector(x_vp); 
    free_vector(x_sub_vp);
    free_vector(u_vp); 
    free_vector(centered_x_vp);
    free_vector(ones_vp);
    free_vector(temp_vp);

    free_matrix(I_1_mp);

#ifndef DONT_PLOT_GRAD_BUT_DO_GRAD_ASCENT   
    free_vector(plot_tau_sig_sqr_ratio_grad_vp);
    free_vector(plot_log_likelihood_vp);
    free_vector(log_p_vp);
    free_vector(plot_exp_comp_log_likelihood_vp);
#endif

    free_matrix_vector(S_tilda_inv_mvp);
    free_matrix(S_inv_mp);
    free_matrix(temp_mp);

    if (result == ERROR) 
    {
        return ERROR; 
    }
    else 
    {
        return num_clusters;
    }
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef __cplusplus
}
#endif

